/*
Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
NSW. Australia,  www.forward.com.au
All rights reserved.

Redistribution of the source of this package, with or without
modification, is NOT permitted.

Redistribution of the compiled java .class files of this package, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions in compiled java .class form must reproduce the above
   copyright notice, this list of conditions and the following disclaimer
   in the documentation and/or other materials provided with the distribution.

2. If the software is modified, the compiled java .class distribution must
   include, in the documentation and/or other materials provided with the 
   distribution, a statement that the software has been modified and
   identify the person who was responsible for the modifications.  

3. Neither the name of the author nor Forward Computing and Control may be 
   used to endorse or promote products derived from this software
   without specific prior written permission.

4.  All advertising materials mentioning features or use of this
    software must display the following acknowledgment: 
    "This product includes software developed by Matthew Ford
     and Forward Computing and Control Pty. Ltd. 
    (http://www.forward.com.au)."


THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package au.com.forward.runWST.wstAttributeReader;
import au.com.forward.runWST.wstAttributeReader.language.*;

import java.io.*;
import java.net.URL;
import java.util.*;
import java.util.logging.*;
import antlr.*;
import java.util.regex.*;  // mod mpf

import org.antlr.stringtemplate.*;

import au.com.forward.runWST.maps.*;

/**
 *  A WSTAttributeReader. <br>
 *  This program reads the stringtemplate attributes and populates a Hashtable.
 *  Reads from an html page looking for the first <DL> list
 *  <br>
 *  Calls to this classes methods are <b>NOT</b> thread safe.
 *
 *@author     matthew ford
 *@created    November 26, 2003
 */
public class WSTAttributeReader {

	/**
	 *  logger for this class
	 */
	public final static Logger logger = Logger.getLogger("au.com.forward.runWST.wstAttributeReader.WSTAttributeReader");

	private AttributeHashtable resultHashtable = null;

	private String fileName;

	private StringTemplateErrorListener errorListener = null;

	public static final Pattern attributeNamePattern = Pattern.compile("(\\.|[A-Z]|_|[a-z])(\\w|\\.)*");

	private HtmlMapState mapState = null;


	/**
	 *  Constructor for WSTAttributeReader. <br>
	 *  Calls to this objects methods are not thread safe.
	 *
	 *@param  listener  the error listener (cannot be null)
	 */
	public WSTAttributeReader(StringTemplateErrorListener listener) {
		if (listener == null) {
			throw new IllegalArgumentException("StringTemplateErrorListener cannot be null.");
		}
		errorListener = listener;
	}



	/**
	 *  Load a datafile of attributes from the classpath.
	 *
	 *@param  fileName  classpath to file to load
	 *@return           hashtable of loaded values (null if error loading)
	 */
	public AttributeHashtable loadDataFromClasspath(String fileName) {
		logger.entering(this.getClass().getName(), "loadDataFromClasspath", fileName);
		ClassLoader cl = this.getClass().getClassLoader();
		InputStream is = cl.getResourceAsStream(fileName);
		if (is == null) {
			error("Error: Can't load attribute datafile '" + fileName + "' as resource");
			return null;
		}
		return loadDataFromInputStream(new BufferedInputStream(is), fileName);
	}


	/**
	 *  Load a datafile of attributes from the file.
	 *
	 *@param  fileName  file to be loaded
	 *@return           hashtable of loaded values (null if error loading)
	 */
	public synchronized  AttributeHashtable loadDataFromFile(String fileName) {
		logger.entering(this.getClass().getName(), "loadDataFromFile", fileName);
		BufferedInputStream bis = null;
		try {
			bis = new BufferedInputStream(new FileInputStream(fileName));
		} catch (Throwable t) {
			error("Error: Can't load attribute datafile '" + fileName + "'", t);
			return null;
		}
		return loadDataFromInputStream(bis, fileName);
	}


	/**
	 *  Load a datafile of attributes from open input stream.
	 *
	 *@param  inStream    the stream to read from
	 *@param  streamName  the name of the stream for error messages (cannot be null)
	 *@return           hashtable of loaded values (null if error loading)
	 */
	public synchronized AttributeHashtable loadDataFromInputStream(InputStream inStream, String streamName) {
		logger.entering(this.getClass().getName(), "loadDataFromInputStream", streamName);
    if (streamName == null) {
			throw new IllegalArgumentException("StreamName cannot be null");
		}
		resultHashtable = null;
		this.fileName = streamName;
		mapState = new HtmlMapState(errorListener);
    try {
      HTMLLexer lexer = new HTMLLexer(new InputStreamReader(inStream,"ISO-8859-1"));
      TokenBuffer buffer = new TokenBuffer(lexer);
      HTMLParser parser = new HTMLParser(buffer);
			parser.setMapState(mapState);  // no listner for now
      parser.document();
			resultHashtable = mapState.getTable();
		} catch (Throwable t) {
			if (t instanceof TokenStreamRecognitionException) {
				TokenStreamRecognitionException e = (TokenStreamRecognitionException)t;
			  error("Error reading attributes from '" + fileName + " -- "+e.toString(),null);
			} else { 
				error("Error reading attributes from '" + fileName + "' -- "+t.getMessage(), t);
			}	
			resultHashtable = null;
		} finally {
			try {
				inStream.close();
			} catch (Throwable t) { }	
		}
		return resultHashtable;
	}




	// U T I L I T Y  R O U T I N E S

	/**
	 *  Description of the Method
	 *
	 *@param  msg  Description of the Parameter
	 */
	public void error(String msg) {
		error(msg, null);
	}


	/**
	 *  Description of the Method
	 *
	 *@param  msg  Description of the Parameter
	 */
	public void warning(String msg) {
		if (errorListener != null) {
			errorListener.warning(msg);
		} else {
			System.err.println("WSTAttributeReader: warning: " + msg);
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  msg  Description of the Parameter
	 */
	public void debug(String msg) {
		if (errorListener != null) {
			errorListener.debug(msg);
		} else {
			System.err.println("WSTAttributeReader: debug: " + msg);
		}
	}


	/**
	 *  Description of the Method
	 *
	 *@param  msg  Description of the Parameter
	 *@param  e    Description of the Parameter
	 */
	public void error(String msg, Throwable e) {
		if (errorListener != null) {
			errorListener.error(msg, e);
		} else {
			if (e != null) {
				System.err.println("WSTAttributeReader: error: " + msg + ": " + e.getMessage());
			} else {
				System.err.println("WSTAttributeReader: error: " + msg);
			}
		}
	}

}

