/*
 *  Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsible for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package au.com.forward.webstringtemplate;

import java.io.*;
import java.util.*;
import java.util.logging.*;
import au.com.forward.logging.*;
import au.com.forward.logging.*;
import java.lang.reflect.*;
import java.util.regex.*;

import org.antlr.stringtemplate.language.*;
import org.antlr.stringtemplate.*;
import antlr.*;
import antlr.collections.AST;
import au.com.forward.runWST.maps.AttributeNameComparator;
import au.com.forward.runWST.wstAttributeReader.WSTAttributeReader;
import au.com.forward.webstringtemplate.filters.*;

/**
 *  A WebStringTemplate. <br>
 *  This class holds both the page template and the data map. <p>
 *
 *  See the documentation for details of its use.
 *
 *@author     matthew ford
 *@created    December 26, 2003
 */
public class WebStringTemplate extends StringTemplate implements Map {

	/**
	 *  The Java Logger for this class
	 */
	public final static Logger logger = Logger.getLogger("au.com.forward.webstringtemplate.WebStringTemplate");

	/**
	 *  The version number - 1.5
	 */
	public final static String VERSION = "1.5";
	
	/**
	 *  The version date 22th Jan. 2004
	 */
	public final static String VERSION_DATE = "22th Jan. 2004";
	

	/**
	 *  The default group to attach this template to if none is specified.
	 */
	protected static StringTemplateGroup defaultGroup =
			new WebStringTemplateGroup("defaultGroup", ".");


	/**
	 *  Create a blank template with no pattern and an empty attribute data map
	 *
	 *@param  group  the group to attach this template to. Must be a {@link
	 *      au.com.forward.webstringtemplate.WebStringTemplateGroup}*
	 */
	public WebStringTemplate(WebStringTemplateGroup group) {
		super();
		// replace the StringTemplate default group.
		setGroup(group);
		attributes = new HashMap();
	}


	/**
	 *  Create a blank template with this data map of values
	 *
	 *@param  group          the group to attach this template to. Must be a {@link
	 *      au.com.forward.webstringtemplate.WebStringTemplateGroup}*
	 *@param  initialValues  the data map of values. No check is made that this map
	 *      has valid keys.
	 */
	public WebStringTemplate(WebStringTemplateGroup group, Map initialValues) {
		this(group);
		attributes = initialValues;
		if (attributes == null) {
			attributes = new HashMap();
		}
	}


	/**
	 *  Create an anonymous template with no name, but with a group
	 *
	 *@param  group                      the group to attach this template to. Must
	 *      be a {@link au.com.forward.webstringtemplate.WebStringTemplateGroup}
	 *@param  template                   the template as a String
	 *@throws  IllegalArgumentException  if group null or not a {@link
	 *      au.com.forward.webstringtemplate.WebStringTemplateGroup}
	 */
	public WebStringTemplate(WebStringTemplateGroup group, String template) {
		this(group);
		// sets group to default
		if (group != null) {
			if (!(group instanceof WebStringTemplateGroup)) {
				throw new IllegalArgumentException("Group must be a WebStringTemplateGroup.");
			}
			setGroup(group);
		}
		setTemplate(template);
	}


	/**
	 *  Make the 'to' template look exactly like the 'from' template except for the
	 *  attributes.
	 *
	 *@param  from  the from StringTemplate
	 *@param  to    the to StringTemplate
	 */
	protected void dup(StringTemplate from, StringTemplate to) {
		super.dup(from, to);
		if (to.getAttributes() == null) {
			to.setAttributes(new HashMap());
		}
	}



	/**
	 *  Sets the group of this WebStringTemplate
	 *
	 *@param  group                      The new group. Must be a {@link
	 *      WebStringTemplateGroup}. Cannot be null.
	 *@throws  IllegalArgumentException  if group null or not a {@link
	 *      WebStringTemplateGroup}
	 */
	public void setGroup(StringTemplateGroup group) {
		if (group != null) {
			if (!(group instanceof WebStringTemplateGroup)) {
				throw new IllegalArgumentException("Group must be a WebStringTemplateGroup");
			}
			this.group = group;
		} else {
			throw new IllegalArgumentException("Group cannot be null");
		}
	}

	/**
	 *  Gets the group of this WebStringTemplate. <br>
	 *  Just calles super.getGroup() but casts result to {@link WebStringTemplateGroup} 
	 *
	 *@return  group    a {@link WebStringTemplateGroup}
	 */
	public WebStringTemplateGroup getWebStringTemplateGroup() {
		return (WebStringTemplateGroup)getGroup();
	}

	/**
	 *  Sets the group of this WebStringTemplate
	 *
	 *@param  group         The new group. Cannot be null.
	 *@throws  IllegalArgumentException  if group null
	 */
	public void setWebStringTemplateGroup(WebStringTemplateGroup group) {
		setGroup(group);
	}
	
	/*
	 *  ************* replaced my addAttribute below   mod mpf
	 *  ** Add an attribute for this template.  If you set the same
	 *  attribute more than once, you get a multi-valued attribute.
	 *  If you send in a StringTemplate object as a value, it's
	 *  enclosing instance (where it will inherit values from) is
	 *  set to 'this'.  This would be the normal case, though you
	 *  can set it back to null after this call if you want.
	 *  If you send in a List plus other values to the same
	 *  attribute, they all get flattened into one List of values.
	 *  If you send in an array, it is converted to a List.  Works
	 *  with arrays of objects and arrays of {int,float,double}.
	 *
	 */
	/**
	 *  Set an attribute. <br>
	 *  null values are stored as "" (an empty string).<br>
	 *  If the attribute name ends in .extension, look for assocaited filter. If a
	 *  filter is found then strip the extension and wrap value in the filter, only
	 *  if it not already wrapped.<br>
	 *  names that start with wst. are control attributes and are not check for
	 *  filter extensions.
	 *
	 *@param  name   The attribute name
	 *@param  value  The new attribute value
	 */
	public void setAttribute(String name, Object value) {
		// mod mpf added
		// is this a control attribute
		boolean controlAttribute = name.startsWith(WST_ATTRIBUTE_START);
		if (!controlAttribute) {
			int idx = name.lastIndexOf(".");
			if (idx >= 0) {
				String nameExtension = name.substring(idx, name.length());
				IWSTFilter filter = ((WebStringTemplateGroup) getGroup()).getFilter(nameExtension);
				if (filter != null) {
					// strip extension
					name = name.substring(0, idx);
					value = wrapWithFilter(value, filter);
				}
			}
		}
		// must call rawSetAttribute to check for valid name
		rawSetAttribute(name, value);
	}


	/**
	 *  Set this attribute name with the given value. <br>
	 *  Attributes whose name includes a . cannot be accessed by the template.
	 *
	 *@param  name                        the attribute name
	 *@param  value                       the value. If the value is a
	 *      StringTemplate its enclosingInstance is set to this template.
	 *@throws  NullPointerException       if name is null
	 *@throws  IllegalArguementException  if the name does not match [. | _ |
	 *      alpha][. | _ | alpha | digit]*
	 */
	protected void rawSetAttribute(String name, Object value) {
		if (name == null) {
			throw new NullPointerException("Attribute name cannot be null.");
		}
		if (value == null) {
			// return;
			value = "";
			// mod mpf
		}
		if (value instanceof StringTemplate) {
			((StringTemplate) value).setEnclosingInstance(this);
		}
		// check for valid name
		name = name.trim();
		if (name.length() <= 0) {
			throw new IllegalArgumentException("Attribute name cannot be empty or blank.");
		}
		// only these chars are allowed by the stringtemplate parser
		if (!WSTAttributeReader.attributeNamePattern.matcher(name).matches()) {
			throw new IllegalArgumentException("Attribute name must be letter or . or _ followed by letter or . or _ or digit. Found '" + name + "'");
		}
		attributes.put(name, value);
	}


	/**
	 *  Gets the attribute value from this WebStringTemplate object. <br>
	 *  Just calls {@link #get(StringTemplate,String) get(this,name)}
	 *
	 *@param  name  the attribute key
	 *@return       The attribute value, wrapped in a filter
	 */
	public Object getAttribute(String name) {
		return get(this, name);
	}


	/**
	 *  Resolve an attribute reference. <br>
	 *  It can be in three possible places: <br>
	 *  1. the attribute list for the current template <br>
	 *  2. if self is an embedded template, somebody invoked us possibly with
	 *  arguments--check the argument context <br>
	 *  3. if self is an embedded template, the attribute list for the enclosing
	 *  instance
	 *
	 *@param  self       the StringTemplate to look in first
	 *@param  attribute  the attribute key
	 *@return            The attribute value, wrapped in a filter
	 */
	public Object get(StringTemplate self, String attribute) {
		Object obj = super.get(self, attribute);
		obj = convertToArrayList(obj);
		// wrapped in filter
		return obj;
	}


	/**
	 *  Description of the Field
	 */
	public final static String WST_ATTRIBUTE_START = "wst.";

	/**
	 *  The name of the attibute (wst.bufferSize) that sets the output buffer size
	 *  for this template
	 */
	public final static String OUTPUT_BUFFER_SIZE = WST_ATTRIBUTE_START + "bufferSize";

	/**
	 *  The name of the attribute (wst.outputDataDefinition) that controls writing
	 *  out the attribute map as a HTML data definition map. <br>
	 *  If this attribute exists and is not false, the HTML data definition map
	 *  will be written out instead of the processed template.
	 */
	public final static String OUTPUT_DATA_DEFINITION_FLAG = WST_ATTRIBUTE_START + "outputDataDefinition";

	/**
	 *  The name of the attribute (wst.outputDataDefinition.maxDepth) that sets the
	 *  recursion depth when writing out the HTML data map. <br>
	 *  Only used if {@link au.com.forward.webstringtemplate.WebStringTemplate#OUTPUT_DATA_DEFINITION_FLAG
	 *  wst.outputDataDefinition} is not false.
	 */
	public final static String OUTPUT_MAX_DEPTH = WST_ATTRIBUTE_START + "outputDataDefinition.maxDepth";

	/**
	 *  The name of the attribute (wst.outputDataDefinition.showAll) that controls
	 *  if all attributes should be output. <br>
	 *  If this attribute exists and is not false, the HTML data map will contain
	 *  all the attributes in the map even if they cannot be accessed by the
	 *  template. Only used if {@link au.com.forward.webstringtemplate.WebStringTemplate#OUTPUT_DATA_DEFINITION_FLAG
	 *  wst.outputDataDefinition} is not false.
	 */
	public final static String OUTPUT_SHOW_ALL_FLAG = WST_ATTRIBUTE_START + "outputDataDefinition.showAll";

	/**
	 *  The name of the attribute (wst.outputDataDefinition.showFilters) that
	 *  controls if all the filter class name used for this string should be output
	 *  before the string. <br>
	 *  Only used if {@link au.com.forward.webstringtemplate.WebStringTemplate#OUTPUT_DATA_DEFINITION_FLAG
	 *  wst.outputDataDefinition} is not false.
	 */
	public final static String OUTPUT_SHOW_FILTERS_FLAG = WST_ATTRIBUTE_START + "outputDataDefinition.showFilters";

	/**
	 *  The attribute name ending (.comment) that signifies an attribute comment
	 *  (case insensitive). <br>
	 *  Use this ending to create an attribute name to hold a comment describing
	 *  the attribute.
	 */
	public final static String WST_COMMENT_ENDING = ".comment";

	/**
	 *  The name of the default HTML data map style resource (dataStyle.css). <br>
	 *
	 */
	public final static String DATA_CSS_NAME = "dataStyle.css";

	/**
	 *  The name of the default HTML data map style file (dataStyle.css). <br>
	 *  Used if built-in style resource cannot be found.
	 */
	public final static String DATA_CSS_FILE_PATH = "dataStyle.css";

	private boolean showAll = false;
	private boolean showFilters = false;

	private final static String indentIncrement = "   ";

	/**
	 *  Default recursion level (3) used when writing out the HTML data map.
	 */
	public final static int DEFAULT_MAX_DEPTH = 3;

	/**
	 *  Default buffer size, 8192
	 */
	public final static int DEFAULT_BUFFER_SIZE = 8192;

	private final static Pattern isPattern = Pattern.compile("is[A-Z].*");
	private final static Pattern getPattern = Pattern.compile("get[A-Z].*");


	/**
	 *  Gets the nameForPath attribute of the WebStringTemplate object
	 *
	 *@return    The nameForPath value
	 */
	public String getNameForPath() {
		String name = super.getName();
		// may have file path in it if loaded from class.
		// replace any / or \ with _
		name = name.replace('/', '_');
		name = name.replace('\\', '_');
		return name;
	}


	/**
	 *  un-encode Unicode characters before passing to StringTemplate
	 *
	 *@param  template  The new template value
	 */
	public void setTemplate(String template) {
		template = loadConvert(template);
		super.setTemplate(template);
	}


	/*
	 *  ************** map interface methods ******************
	 */
	/**
	 *  Returns the value to which this template data map maps the specified key.
	 *  Returns <tt>null</tt> if the map contains no mapping for this key.
	 *
	 *@param  key                  the String key whose associated value is to be
	 *      returned.
	 *@return                      the value to which this map maps the specified
	 *      key, or <tt>null</tt> if the map contains no mapping for this key or if
	 *      the key is null.
	 *@throws  ClassCastException  if the key is not a String.
	 *@see                         #containsKey(Object)
	 */
	public Object get(Object key) {
		if (key == null) {
			return null;
		}
		String stringKey = (String) key;
		return convertToArrayList(attributes.get(stringKey));
		// wrapped in filter
	}


	/**
	 *  Associates the specified value with the specified attribute name in this
	 *  template data map. <br>
	 *  If the map previously contained a mapping for this attribute name, the old
	 *  value is replaced by the specified value.<p>
	 *
	 *  If the attribute name ends in .extension, look for assocaited filter. If a
	 *  filter is found then strip the extension and wrap value in the filter, only
	 *  if it not already wrapped.<br>
	 *  names that start with wst. are control attributes and are not check for
	 *  filter extensions.
	 *
	 *@param  attributeName              the name of this of this value.
	 *@param  value                      value to be associated with this attribute
	 *      name.
	 *@return                            previous value associated with this
	 *      attribute name, or <tt>null</tt> if there was no mapping for key.
	 *@throws  ClassCastException        if the attributeName is not a String.
	 *@throws  IllegalArgumentException  if the key is not a valid attributeName
	 *      (i.e [. | _ | alpha][. | _ | alpha | digit]* )
	 */
	public Object put(Object attributeName, Object value) {
		String name = (String) attributeName;
		boolean controlAttribute = name.startsWith(WST_ATTRIBUTE_START);
		if (!controlAttribute) {
			int idx = name.lastIndexOf(".");
			if (idx >= 0) {
				String nameExtension = name.substring(idx, name.length());
				IWSTFilter filter = ((WebStringTemplateGroup) getGroup()).getFilter(nameExtension);
				if (filter != null) {
					// strip extension
					name = name.substring(0, idx);
					value = wrapWithFilter(value, filter);
				}
			}
		}
		Object rtn = get(name);
		// must call rawSetAttribute to check for valid name
		rawSetAttribute(name, value);
		return rtn;
	}


	/**
	 *  Returns the number of key-value mappings in this template data map. If the
	 *  map contains more than <tt>Integer.MAX_VALUE</tt> elements, returns <tt>
	 *  Integer.MAX_VALUE</tt> .
	 *
	 *@return    the number of key-value mappings in this map.
	 */
	public int size() {
		return attributes.size();
	}


	/**
	 *  Returns <tt>true</tt> if this template data map contains no key-value
	 *  mappings.
	 *
	 *@return    <tt>true</tt> if this map contains no key-value mappings.
	 */
	public boolean isEmpty() {
		return attributes.isEmpty();
	}


	/**
	 *  Returns <tt>true</tt> if this template data map contains a mapping for the
	 *  specified key.
	 *
	 *@param  key                  the String key whose presence in this map is to
	 *      be tested.
	 *@return                      <tt>true</tt> if this map contains a mapping for
	 *      the specified key.
	 *@throws  ClassCastException  if the key is not a String.
	 */
	public boolean containsKey(Object key) {
		if (key == null) {
			return false;
		}
		String stringKey = (String) key;
		return attributes.containsKey(stringKey);
	}


	/**
	 *  Returns <tt>true</tt> if this template data map maps one or more keys to
	 *  the specified value. This operation will probably require time linear in
	 *  the map size for most implementations.
	 *
	 *@param  value                  value whose presence in this map is to be
	 *      tested.
	 *@return                        <tt>true</tt> if this map maps one or more
	 *      keys to the specified value.
	 *@throws  NullPointerException  if the value is <tt>null</tt> .
	 */
	public boolean containsValue(Object value) {
		if (value == null) {
			throw new NullPointerException("Value cannot be null.");
		}
		return attributes.containsValue(value);
	}


	// Modification Operations


	/**
	 *  Removes the mapping for this key from this template data map if it is
	 *  present. <p>
	 *
	 *  Returns the value to which the map previously associated the key, or <tt>
	 *  null</tt> if the map contained no mapping for this key. The map will not
	 *  contain a mapping for the specified key once the call returns.
	 *
	 *@param  key                    the String key whose mapping is to be removed
	 *      from the map.
	 *@return                        previous value associated with specified key,
	 *      or <tt>null</tt> if there was no mapping for key.
	 *@throws  ClassCastException    if the key is not a String.
	 *@throws  NullPointerException  if the key is <tt>null</tt> .
	 */
	public Object remove(Object key) {
		if (key == null) {
			throw new NullPointerException("Key cannot be null.");
		}
		String stringKey = (String) key;

		return attributes.remove(stringKey);
	}


	// Bulk Operations

	/**
	 *  Copies all of the mappings from the specified map to this map. This
	 *  operation is not supported on template data maps.
	 *
	 *@param  t                               map to put
	 *@throws  UnsupportedOperationException  if the <tt>putAll</tt> method is
	 *      called.
	 */
	public void putAll(Map t) {
		throw new UnsupportedOperationException("putAll(map) not supported for template attributes.");
	}


	/**
	 *  Removes all mappings from this template data map.
	 */
	public void clear() {
		attributes.clear();
	}


	// Views

	/**
	 *  Returns a set view of the keys contained in this template data map. The set
	 *  is backed by the map, so changes to the map are reflected in the set, and
	 *  vice-versa. If the map is modified while an iteration over the set is in
	 *  progress, the results of the iteration are undefined. The set supports
	 *  element removal, which removes the corresponding mapping from the map, via
	 *  the <tt>Iterator.remove</tt> , <tt>Set.remove</tt> , <tt>removeAll</tt>
	 *  <tt>retainAll</tt> , and <tt>clear</tt> operations. It does not support the
	 *  add or <tt>addAll</tt> operations.
	 *
	 *@return    a set view of the keys contained in this map.
	 */
	public Set keySet() {
		return attributes.keySet();
	}


	/**
	 *  Returns a collection view of the values contained in this template data
	 *  map. The collection is backed by the map, so changes to the map are
	 *  reflected in the collection, and vice-versa. If the map is modified while
	 *  an iteration over the collection is in progress, the results of the
	 *  iteration are undefined. The collection supports element removal, which
	 *  removes the corresponding mapping from the map, via the <tt>Iterator.remove
	 *  </tt>, <tt>Collection.remove</tt> , <tt>removeAll</tt> , <tt>retainAll</tt>
	 *  and <tt>clear</tt> operations. It does not support the add or <tt>addAll
	 *  </tt> operations.
	 *
	 *@return    a collection view of the values contained in this map.
	 */
	public Collection values() {
		return attributes.values();
	}


	/**
	 *  Returns a set view of the mappings contained in this template data map.
	 *  Each element in the returned set is a {@link java.util.Map.Entry}. The set
	 *  is backed by the map, so changes to the map are reflected in the set, and
	 *  vice-versa. If the map is modified while an iteration over the set is in
	 *  progress, the results of the iteration are undefined. The set supports
	 *  element removal, which removes the corresponding mapping from the map, via
	 *  the <tt>Iterator.remove</tt> , <tt>Set.remove</tt> , <tt>removeAll</tt> ,
	 *  <tt>retainAll</tt> and <tt>clear</tt> operations. It does not support the
	 *  <tt>add</tt> or <tt>addAll</tt> operations.
	 *
	 *@return    a set view of the mappings contained in this map.
	 */
	public Set entrySet() {
		return attributes.entrySet();
	}


	/*
	 *  ************************ end of Map interface methods ***********************
	 */
	/**
	 *  Tests if an object is not false. <br>
	 *  Returns false for a null object, obj.toString()=="false" or
	 *  obj.toString()=="" or obj.toString()=="0" (zero)<br>
	 *  All other objects return true
	 *
	 *@param  a  the object to test
	 *@return    true if the object is not false
	 */
	public boolean testObjectTrue(Object a) {
		boolean value;
		if (a == null) {
			value = false;
		} else {
			IWSTFilter filter = null;
			Object trueObj = a;
			while (trueObj instanceof IWSTFilter) {
				filter = (IWSTFilter) trueObj;
				trueObj = filter.getObject();
			}
			if (filter instanceof WSTMissingFilter) {
				return false;
			}
			// else

			// if (a!=null) {
			// try toString == "false" or ""
			try {
				String result = convertToString(a);
				// this handles objects whose toString() returns null
				// also handles ints etc
				value = !(result.equals("false") || result.equals("")
						 || result.equals("0"));
			} catch (Throwable t) {
				// some error calling toString() so != "false"
				value = true;
			}
		}
		return value;
	}


	/**
	 *  Wrap an object with the given filter, but only if the object is not already
	 *  wrapped.
	 *
	 *@param  value   the object to wrap
	 *@param  filter  the fiter to wrap it in, if it is not already wrapped. If
	 *      null use the filter returned by {@link
	 *      WebStringTemplateGroup#getDefaultFilter()}.
	 *@return         the wrapped object
	 */
	public Object wrapWithFilter(Object value, IWSTFilter filter) {
		if (value instanceof IWSTFilter) {
			//all already filtered
			return value;
		}
		if (filter == null) {
			// no filter use default
			filter = ((WebStringTemplateGroup) getGroup()).getDefaultFilter();
		}
		// else wrap in new cloned filter
		IWSTFilter newFilter = filter.newInstance();
		newFilter.setObject(value);
		newFilter.setPath(filter.getPath());
		return newFilter;
	}


	/**
	 *  Convert to string and trim leading and trailing spaces
	 *
	 *@param  obj  the object to convert
	 *@return      ""+obj or a string containing the error that occured while
	 *      trying to convert obj to a string.
	 */
	public String convertToString(Object obj) {
		//logger.entering(this.getClass().getName(),"convertToString",obj);
		String rtn = "";
		if (obj != null) {
			// if null leave rtn as ""
			Object trueObj = obj;
			IWSTFilter filter = null;
			while (trueObj instanceof IWSTFilter) {
				filter = (IWSTFilter) trueObj;
				trueObj = filter.getObject();
			}

			// try map.toString()
			if (trueObj instanceof Map) {
				// is there a "" key
				Object mapString = ((Map)trueObj).get("");
				if (mapString != null) {
					obj = wrapWithFilter(mapString,filter);
				} 
			}	
			try {
				rtn = "" + obj;
			} catch (Throwable t) {
				logger.log(Level.SEVERE, "Error converting " + obj.getClass().getName() + " to String", t);
				rtn = ("Error converting " + obj.getClass().getName() + " to String\n" + RobustFormatter.toString(t));
				//throw new RuntimeException("Error calling toString() on "+obj.getClass().getName(),t);
			}
		}
		//logger.exiting(this.getClass().getName(),"convertToString",rtn);
		return rtn.trim();
	}


	/**
	 *  Convert value to arrayList if needed. <br>
	 *  In any case wrap in defaultFilter in not already wrapped in a filter
	 *
	 *@param  value  the object to convert
	 *@return        the object converted to an ArrayList if it was an array or
	 *      Enumeration or Iterator or Collection. The return object is always
	 *      wrapped.<br>
	 *      If the input value was wrapped, the return and all its list elements
	 *      are also wrapped in the same filter, unless the element was already
	 *      wrapped, else <br>
	 *      if the input was not wrapped then the return and all its list elements
	 *      are also wrapped in the the filter returned by {@link
	 *      WebStringTemplateGroup#getDefaultFilter()}, , unless the element was
	 *      already wrapped.
	 */
	public Object convertToArrayList(Object value) {
		logger.entering(this.getClass().getName(), "convertToArrayList", value);
		IWSTFilter filter = null;
		if (value instanceof IWSTFilter) {
			while (value instanceof IWSTFilter) {
				// unwrap it first
				filter = (IWSTFilter) value;
				value = filter.getObject();
			}
		} else {
			StringTemplateGroup group = getGroup();
			if (group instanceof WebStringTemplateGroup) {
				filter = ((WebStringTemplateGroup) group).getDefaultFilter();
			}
		}
		if (value instanceof Object[]) {
			Object[] list = (Object[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				Object elem = list[i];
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;

		} else if (value instanceof boolean[]) {
			boolean[] list = (boolean[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				boolean belem = list[i];
				Object elem = new Boolean(belem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;

		} else if (value instanceof byte[]) {
			byte[] list = (byte[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				byte belem = list[i];
				Object elem = new Byte(belem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;

		} else if (value instanceof char[]) {
			char[] list = (char[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				char celem = list[i];
				Object elem = new Character(celem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;

		} else if (value instanceof short[]) {
			short[] list = (short[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				short selem = list[i];
				Object elem = new Short(selem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;

		} else if (value instanceof int[]) {
			int[] list = (int[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				int ielem = list[i];
				Object elem = new Integer(ielem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;
		} else if (value instanceof float[]) {
			float[] list = (float[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				float felem = list[i];
				Object elem = new Float(felem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;
		} else if (value instanceof double[]) {
			double[] list = (double[]) value;
			List v = new ArrayList(list.length);
			for (int i = 0; i < list.length; i++) {
				double delem = list[i];
				Object elem = new Double(delem);
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;

		} else if (value instanceof Enumeration) {
			// put enumerations into arraylist
			List v = new ArrayList();
			Enumeration e = (Enumeration) value;
			for (; e.hasMoreElements(); ) {
				Object elem = e.nextElement();
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;
		} else if ((value instanceof Iterator) || (value instanceof Collection)) {
			// put enumerations into arraylist
			List v = new ArrayList();
			Iterator it = null;
			if (value instanceof Iterator) {
				it = (Iterator) value;
			} else if (value instanceof Collection) {
				it = ((Collection) value).iterator();
			} else {
				// ??
				value = wrapWithFilter(value, filter);
				// only if filter not null
				logger.exiting(this.getClass().getName(), "convertToArrayList", value);
				return value;
			}
			for (; it.hasNext(); ) {
				Object elem = it.next();
				elem = wrapWithFilter(elem, filter);
				// only if filter not null
				v.add(elem);
			}
			value = v;
		}
		// else
		value = wrapWithFilter(value, filter);
		// only if filter not null
		logger.exiting(this.getClass().getName(), "convertToArrayList", value);
		return value;
		// note there is a return from Interator above also.
	}



	/**
	 *  Adds an attribute for this template data map. <br>
	 *  If you add the same attribute more than once, you get a multi-valued
	 *  attribute.<br>
	 *  If the attribute name ends in .extension, look for assocaited filter. If a
	 *  filter is found then strip the extension and wrap value in the filter, only
	 *  if it not already wrapped.<br>
	 *  names that start with wst. are control attributes and are not check for
	 *  filter extensions. If you send in a StringTemplate object as a value, it's
	 *  enclosing instance (where it will inherit values from) is set to 'this'.
	 *  This would be the normal case, though you can set it back to null after
	 *  this call if you want. <br>
	 *  If you send in a List plus other values to the same attribute, they all get
	 *  flattened into one List of values. <br>
	 *  If you send in an array, it is converted to a List. Works with arrays of
	 *  objects and arrays of {int,short,float,double,byte,char,boolean} and
	 *  Enumeration, Iterator and Collection.
	 *
	 *@param  name                        The name of this attribute
	 *@param  value                       the value associated with this attribute
	 *      name. Null values are saved as ""
	 *@throws  NullPointerException       if name is null
	 *@throws  IllegalArguementException  if the name does not match [. | _ |
	 *      alpha][. | _ | alpha | digit]*
	 */
	public void addAttribute(String name, Object value) {
		if (value == null) {
			value = "";
		}
		boolean controlAttribute = name.startsWith(WST_ATTRIBUTE_START);
		if (!controlAttribute) {
			int idx = name.lastIndexOf(".");
			if (idx >= 0) {
				String nameExtension = name.substring(idx, name.length());
				IWSTFilter filter = ((WebStringTemplateGroup) getGroup()).getFilter(nameExtension);
				if (filter != null) {
					// strip extension
					name = name.substring(0, idx);
					value = wrapWithFilter(value, filter);
				}
			}
		}

		// get exactly in this scope (no enclosing)
		Object o = attributes.get(name);
		// returns raw object may not be arrayList and may not be wrapped


		if (o != null) {
			o = convertToArrayList(o);
			attributes.put(name, o);
			// put it back now that it is a wrapped list
			Object trueO = o;
			IWSTFilter oFilter = null;
			while (trueO instanceof IWSTFilter) {
				// unwrap it first
				oFilter = (IWSTFilter) trueO;
				trueO = oFilter.getObject();
			}
			// OK here trueO is the actual object

			// not trueO here as if o != null then something has been already stored but may have been a null wrapped
			value = convertToArrayList(value);
			// wraps in defaultFilter if not already wrapped

			Object trueValue = value;
			IWSTFilter valueFilter = null;
			while (trueValue instanceof IWSTFilter) {
				// unwrap it first
				valueFilter = (IWSTFilter) trueValue;
				trueValue = valueFilter.getObject();
			}
			// and trueValue is the actual value

			// it's a multi-value attribute
			// if input value == null just clear existing attribute
			// if existing attribute (o) == null just add new value
			//System.out.println("exists: "+name+"="+o);
			ArrayList v = null;
			if (trueO instanceof ArrayList) {
				// already a List
				v = (ArrayList) trueO;
				if (trueValue instanceof List) {
					// flatten incoming list into existing
					// (do nothing if same List to avoid trouble)
					List v2 = (List) trueValue;
					for (int i = 0; v != v2 && i < v2.size(); i++) {
						// System.out.println("flattening "+name+"["+i+"]="+v2.elementAt(i)+" into existing value");
						v.add(wrapWithFilter(v2.get(i), valueFilter));
					}
				} else {
					v.add(wrapWithFilter(value, oFilter));
					// else just add it all wrap it in existing filter if not already wrapped
				}
			} else {
				// second attribute, and existing is not an array list
				// must start a new ArrayList to hold both
				// make list to hold multiple values
				v = new ArrayList();
				// make it point to list now
				// must call rawSetAttribute to check for valid name
				rawSetAttribute(name, v);
				// add new list
				// add previous single-valued attribute
				v.add(o);
				// wrapped
				if (trueValue instanceof List) {
					// flatten incoming list into existing
					List v2 = (List) trueValue;
					for (int i = 0; i < v2.size(); i++) {
						v.add(wrapWithFilter(v2.get(i), valueFilter));
					}
				} else {
					v.add(wrapWithFilter(value, oFilter));
					//v.add(value);
					// add to new list just added to map
				}

			}
		} else {
			// must call rawSetAttribute to check for valid name
			rawSetAttribute(name, value);
		}
	}


	/**
	 *  Replaces non-ascii chars with unicode escapes and replaces </wst> in data output with \u003C/wst>, wrapps the whole in <wst> </wst> .<br>
	 Handles null input. <br>
	 *  This allows the data to be read back in exactly as it would appear in the
	 *  template.
	 *
	 *@param  input  the string to escape
	 *@return        the escaped string, or null if input was null
	 */
	public static String dataEscape(String input) {
		if (input == null) {
			return null;
		}
		input = saveConvert(input);
		// replace any </wst> , ignoring case, in the data with
		// \u003C/wst>  
		StringBuffer lowerCaseInput = new StringBuffer(input.toLowerCase());
		StringBuffer output = new StringBuffer(input);
		int idx = 0;
		while ((idx = lowerCaseInput.indexOf("</wst>",idx)) >= 0) {
			lowerCaseInput.setCharAt(idx,'\\');
			output.setCharAt(idx,'\\');
			lowerCaseInput.insert(idx+1,"u003C");
			output.insert(idx+1,"u003C");
		}	
		return "<wst>"+output+"</wst>";
	}


	/**
	 *  Converts unicodes to encoded &#92;uxxxx matches loadConvert defined
	 *  here and called in HtmlMapState. <br>
	 *  Does not convert \r \n \t or \f
	 *
	 *@param  theString  Description of the Parameter
	 *@return            Description of the Return Value
	 */
	private static String saveConvert(String theString) {
		int len = theString.length();
		StringBuffer outBuffer = new StringBuffer(len * 2);

		for (int x = 0; x < len; x++) {
			char aChar = theString.charAt(x);
			switch (aChar) {
							case '\\':
								outBuffer.append('\\');
								outBuffer.append('\\');
								break;
							case '\t':
								outBuffer.append('\t');
								break;
							case '\n':
								outBuffer.append('\n');
								break;
							case '\r':
								outBuffer.append('\r');
								break;
							case '\f':
								outBuffer.append('\f');
								break;
							default:
								if ((aChar < 0x0020) || (aChar > 0x007e)) {
									outBuffer.append('\\');
									outBuffer.append('u');
									outBuffer.append(toHex((aChar >> 12) & 0xF));
									outBuffer.append(toHex((aChar >> 8) & 0xF));
									outBuffer.append(toHex((aChar >> 4) & 0xF));
									outBuffer.append(toHex(aChar & 0xF));
								} else {
									outBuffer.append(aChar);
								}
			}
		}
		return outBuffer.toString();
	}


	/**
	 *  Convert a nibble to a hex character
	 *
	 *@param  nibble  the nibble to convert.
	 *@return         Description of the Return Value
	 */
	private static char toHex(int nibble) {
		return hexDigit[(nibble & 0xF)];
	}


	/**
	 *  A table of hex digits
	 */
	private final static char[] hexDigit = {
			'0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'
			};


	/**
	 *  Converts encoded &#92;uxxxx to unicode chars matches saveConvert in
	 *  WebStringTemplate
	 *
	 *@param  theString  the string to convert
	 *@return            the converted string
	 */
	public static String loadConvert(String theString) {
		char aChar;
		int len = theString.length();
		StringBuffer outBuffer = new StringBuffer(len);

		for (int x = 0; x < len; ) {
			aChar = theString.charAt(x++);
			if (aChar == '\\') { // must have following escape char 
				aChar = theString.charAt(x++);
				if (aChar == 'u') { // must have dddd following \ u
					// Read the xxxx
					int value = 0;
					for (int i = 0; i < 4; i++) {
						aChar = theString.charAt(x++);
						switch (aChar) {
										case '0':
										case '1':
										case '2':
										case '3':
										case '4':
										case '5':
										case '6':
										case '7':
										case '8':
										case '9':
											value = (value << 4) + aChar - '0';
											break;
										case 'a':
										case 'b':
										case 'c':
										case 'd':
										case 'e':
										case 'f':
											value = (value << 4) + 10 + aChar - 'a';
											break;
										case 'A':
										case 'B':
										case 'C':
										case 'D':
										case 'E':
										case 'F':
											value = (value << 4) + 10 + aChar - 'A';
											break;
										default:
											throw new IllegalArgumentException(
													"Malformed \\uxxxx encoding.");
						}
					}
					outBuffer.append((char) value);
				} else {
					outBuffer.append(aChar);
					// handles \\ and strips \ from all other escapes
				}
			} else {
				outBuffer.append(aChar);
			}
		}
		return outBuffer.toString();
	}


	/**
	 *  Returns the internal style sheet string.
	 *
	 *@return    The CSS as a string or null if the resource is not found.
	 */
	protected String styleSheetString() {
		String rtn = "";
		try {
			ClassLoader cl = this.getClass().getClassLoader();
			InputStream is = cl.getResourceAsStream(DATA_CSS_NAME);
			if (is == null) {
				return null;
			}
			int NO_OF_CHARS = 1000;
			char[] input = new char[NO_OF_CHARS];
			int count = 0;
			Reader reader = new InputStreamReader(new BufferedInputStream(is, 1000), "ISO-8859-1");
			while (count > -1) {
				rtn += new String(input, 0, count);
				count = reader.read(input, 0, NO_OF_CHARS);
			}
		} catch (Throwable t) {
			warning("Error reading " + DATA_CSS_NAME + " " + t.getMessage());
			return null;
		}
		return rtn;
	}


	/**
	 *  Outputs the template data map as a HTML data map. <br>
	 *  Any < in the data are escaped with &amp;shy; to ensure the data can be read
	 *  back in exactly as output.<p>
	 *
	 *  The output is controlled by the {@link
	 *  WebStringTemplate#OUTPUT_SHOW_ALL_FLAG wst.outputDataDefinition.showAll}
	 *  and {@link WebStringTemplate#OUTPUT_MAX_DEPTH
	 *  wst.outputDataDefinition.maxDepth} settings.
	 *
	 *@param  table           the template data map to output. (Should work for any
	 *      map.)
	 *@param  styleSheetPath  the style sheet path for an external stylesheet or
	 *      null to embed the default one in the output file.
	 *@param  title           the title of the HTML output page
	 *@return                 the HTML output
	 */
	public String attributeHashtableToHTML(Map table, String title, String styleSheetPath) {
		StringWriter buf = new StringWriter();
		try {
			attributeHashtableToHTML(table, title, styleSheetPath, DEFAULT_MAX_DEPTH, buf);
		} catch (IOException io) {
			warning("Got IOException writing to StringWriter????");
		}
		return buf.toString();
	}


	/**
	 *  Outputs the template data map as a HTML data map. <br>
	 *  Any < in the data are escaped with &amp;shy; to ensure the data can be read
	 *  back in exactly as output.<p>
	 *
	 *  The output is controlled by the {@link
	 *  WebStringTemplate#OUTPUT_SHOW_ALL_FLAG wst.outputDataDefinition.showAll}
	 *  setting.
	 *
	 *@param  table            the template data map to output. (Should work for
	 *      any map.)
	 *@param  styleSheetPath   the style sheet path for an external stylesheet or
	 *      null to embed the default one in the output file.
	 *@param  title            the title of the HTML output page
	 *@param  maxLevel         the maximum depth to recurse to when outputting.
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	public void attributeHashtableToHTML(Map table, String title, String styleSheetPath, int maxLevel, Writer out)
			 throws IOException {
		int level = maxLevel;
		out.write("<HTML>\n<HEAD>\n");
		// encoding set by final writer but all the output from here is ISO-8859-1 so specify it here anyway.
		out.write("<meta http-equiv=\"Content-Type\" content=\"text/html; charset=ISO-8859-1\" >\n");
		if ((styleSheetPath != null) && (styleSheetPath.trim().length() > 0)) {
			out.write("<LINK REL=\"StyleSheet\" HREF=\"" + styleSheetPath + "\" TYPE=\"text/css\"/>\n");
		} else {
			String styleSheetStr = styleSheetString();
			if (styleSheetStr != null) {
				out.write("<style type=\"text/css\">");
				out.write(styleSheetStr);
				out.write("</style>");
			} else {
				out.write("<!-- could not find default resource stylesheet '" + DATA_CSS_NAME + "'\n" +
						"     in classpath '" + System.getProperty("java.class.path") + "'\n" +
						"     using default stylesheet file name -->\n");
				out.write("<LINK REL=\"StyleSheet\" HREF=\"" + DATA_CSS_FILE_PATH + "\" TYPE=\"text/css\"/>\n");
			}
		}
		if (title != null) {
			out.write("<TITLE>" + title + " Data Interface Definition</TITLE>\n");
		}
		out.write("</HEAD>\n<BODY>\n");
		if (title != null) {
			out.write("<H1>" + title + " Data Interface Definition</H1>\n<HR/>");
		}
		if (table != null) {
			mapToHTML(table, "", level, out);
		}
		if (showAll && showFilters) {
			// show the filter map also
			out.write("<HR>\n<H2>Filter Map</H2>\n");
			Object[][] filters = ((WebStringTemplateGroup) getGroup()).getAllFilters();
			// always at least the default filter
			out.write("<table border=\"3\"><tr><td align=\"center\"><b>nameExtension</b></td><td align=\"center\"><b>filter</b></td></tr>\n");
			for (int i = 0; i < filters.length; i++) {
				Object[] row = filters[i];
				out.write("<tr><td align=\"center\">");
				if (((String) row[0]).length() == 0) {
					out.write("<i>default</i>");
				} else {
					out.write("." + row[0]);
				}
				out.write("</td><td>"
						 + ((row[1] == null) ? "null" : row[1].getClass().getName()) + "</td></tr>\n");
			}
			out.write("</table>\n");
		}

		out.write("</BODY>\n</HTML>\n");
	}



	/**
	 *  This switch method handles maps, lists and objects
	 *
	 *@param  obj              the object to process
	 *@param  indent           the current indent
	 *@param  level            the current level
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	private void processObjectHTML(Object obj, String indent, int level, Writer out) throws IOException {
		// level++; don't increment level here only increment in toHTML methods
		logger.entering(this.getClass().getName(), "processObjectHTML", new Object[]{obj, indent, new Integer(level)});
		indent += indentIncrement;
		obj = convertToArrayList(obj);
		// wrapp in default filter if now already wrapped.
		Object trueObj = obj;
		IWSTFilter filter = null;
		while (trueObj instanceof IWSTFilter) {
			filter = (IWSTFilter) trueObj;
			trueObj = filter.getObject();
		}
		if (trueObj instanceof List) {
			listToHTML(obj, indent, level, out);

		} else if (trueObj instanceof Map) {
			// check for "" key and output it
			Object mapKey = ((Map)trueObj).get("");
			if (mapKey != null) {
				htmlDataObjectToString( wrapWithFilter(mapKey,filter), out);
      }
			mapToHTML(obj, indent, level, out);

		} else if (trueObj == null) {
			return;
			// don't add anything to the buffer
		} else {
			objectToHTML(obj, indent, level, out);
			//return new StringBuffer("" + obj);
		}
		logger.exiting(this.getClass().getName(), "processObjectHTML");
	}


	/**
	 *  Convert this object to a string and add it to the return buffer. <br>
	 *  If not filtered apply default filter.
	 *
	 *@param  obj              Description of the Parameter
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	private void htmlDataObjectToString(Object obj, Writer out) throws IOException {
		logger.entering(this.getClass().getName(), "htmlDataObjectToString", new Object[]{obj});
		Object trueObj = obj;
		IWSTFilter filter = null;
		String filterName = "";
		while (trueObj instanceof IWSTFilter) {
			filter = (IWSTFilter) trueObj;
			filterName += " [" + filter.getClass().getName();
			trueObj = filter.getObject();
		}
		if (filter == null) {
			filter = ((WebStringTemplateGroup) getGroup()).getDefaultFilter();
			obj = wrapWithFilter(trueObj, filter);
			//filterName = " [default: "+filter.getClass().getName()+"] ";
			filterName = " [" + filter.getClass().getName() + "] ";
		} else {
			filterName += "] ";
			// close it off
		}
		if (!showFilters) {
			// clear it out
			// but still wrapped if no filter specified
			filterName = "";
		}

		if (trueObj == null) {
			out.write(filterName);
			return;
		}
		//else {
		String className = trueObj.getClass().getName();
		if (className.startsWith("org.antlr.stringtemplate")
				 || className.startsWith("au.com.forward.webstringtemplate")) {
			// prevent possible loop		 
			out.write(filterName);
			out.write(className);
			return;
		}
		// else
		String str = convertToString(obj);
		// calls toString() if available
		str = dataEscape(str);
		// add <wst>.. </wst> to escape any list tags in the data
		out.write(filterName);
		out.write(str);
		logger.exiting(this.getClass().getName(), "htmlDataObjectToString");
	}


	/**
	 *  Gets the keys of a map as a String[]
	 *
	 *@param  map  the map object
	 *@return      the String[] of it keys
	 */
	private String[] getMapAsStringArrayOfKeys(Map map) {
		if (map.size() == 0) {
			return new String[0];
		}
		Set keys = map.keySet();
		String[] keyArray = (String[]) keys.toArray(new String[0]);
		// sort it
		Arrays.sort(keyArray, keyComparator);
		return keyArray;
	}


	/**
	 *  Output a map in HTML data map format
	 *
	 *@param  mapObj           the map to output
	 *@param  indent           the current indent
	 *@param  level            the current level
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	private void mapToHTML(Object mapObj, String indent, int level, Writer out) throws IOException {
		logger.entering(this.getClass().getName(), "mapToHTML", new Object[]{mapObj, indent, new Integer(level)});
		Object map = mapObj;
		IWSTFilter mapFilter = ((WebStringTemplateGroup) getGroup()).getDefaultFilter();
		if (map instanceof IWSTFilter) {
			while (map instanceof IWSTFilter) {
				mapFilter = (IWSTFilter) map;
				map = mapFilter.getObject();
				if (logger.isLoggable(Level.FINER)) {
					logger.finer("map is filtered with " + mapFilter.getClass().getName());
				}
			}
		}
		level--;
		String[] keys = getMapAsStringArrayOfKeys((Map) map);
		if (keys.length > 0) {
			out.write(indent + "<DL>\n");
			for (int i = 0; i < keys.length; i++) {
				String name = (String) keys[i];
				logger.finer("processing '" + name + "'");
				if (skipThisAttibute(name)) {
					continue;
				}
				out.write(indent + indentIncrement + "<DT>" + name + "</DT>\n");
				out.write(indent + indentIncrement + "<DD>");
				Object obj = ((Map) map).get(name);
				Object trueObj = obj;
				while (trueObj instanceof IWSTFilter) {
					trueObj = ((IWSTFilter) trueObj).getObject();
				}
				if ((trueObj instanceof Map) || (trueObj instanceof List) || (trueObj instanceof Enumeration)) {
					out.write("\n");
				}
				obj = wrapWithFilter(obj, mapFilter);
				// only if filter not null and obj is not already filtered
				if (level < 0) {
					htmlDataObjectToString(obj, out);
					// just toString it
				} else {
					processObjectHTML(obj, indent + indentIncrement, level, out);
				}
				if ((trueObj instanceof Map) || (trueObj instanceof List) || (trueObj instanceof Enumeration)) {
					out.write("\n" + indent + indentIncrement + "</DD>\n");
				} else {
					out.write("</DD>\n");
				}
			}
			out.write(indent + "</DL>\n");
		} else {	
			logger.finer("mapToHTML empty map");
		}
		logger.exiting(this.getClass().getName(), "mapToHTML");
	}


	/**
	 *  Output a list in HTML data map format
	 *
	 *@param  listObj          the list to output
	 *@param  indent           the current indent
	 *@param  level            the current level
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	private void listToHTML(Object listObj, String indent, int level, Writer out) throws IOException {
		logger.entering(this.getClass().getName(), "listToHTML", new Object[]{listObj, indent, new Integer(level)});
		Object list = listObj;
		IWSTFilter listFilter = null;
		if (list instanceof IWSTFilter) {
			while (list instanceof IWSTFilter) {
				listFilter = (IWSTFilter) list;
				list = listFilter.getObject();
				if (logger.isLoggable(Level.FINER)) {
					logger.finer("map is filtered with " + listFilter.getClass().getName());
				}
			}
		} else {
			listFilter = ((WebStringTemplateGroup) getGroup()).getDefaultFilter();
		}
		level--;
		if ( ((List) list).size() > 0) {
			out.write(indent + "<OL>\n");
			for (int i = 0; i < ((List) list).size(); i++) {
				Object obj = ((List) list).get(i);
				obj = wrapWithFilter(obj, listFilter);
				// only if filter not null and obj is not already filtered
				out.write(indent + indentIncrement + "<LI>");
				if (level < 0) {
					htmlDataObjectToString(obj, out);
					// just toString it
				} else {
					processObjectHTML(obj, indent + indentIncrement, level, out);
				}
				out.write("</LI>\n");
			}
			out.write(indent + "</OL>\n");
		} else {	
			logger.finer("listToHTML List is empty");
		}	
		logger.exiting(this.getClass().getName(), "listToHTML");
	}


	/**
	 *  Should this method be skipped when doing object introspection. <br>
	 *  Skip getOutputStream(), getInputStream(), getWriter(), getReader() and getClass().
	 *
	 *@param  methodName  the name of the method
	 *@return             true if this method is to be skipped
	 */
	public boolean skipThisMethod(String methodName) {
		if (methodName.equals("getOutputStream") ||
				methodName.equals("getInputStream") ||
				methodName.equals("getWriter") ||
				methodName.equals("getReader") ||
				methodName.equals("getClass")) {
			return true;
		}
		// else
		return false;
	}


	/**
	 *  Should this class be skipped when output the values of the data map. <br>
	 *  Skip javax.servlet.* and org.apache.* except
	 *  org.apache.taglibs.standard.tag.common.sql.ResultImpl
	 *
	 *@param  className  the string class name of this object.
	 *@return            true if this object should be skipped (not output)
	 */
	public boolean skipThisClass(String className) {
		if (showAll) {
			return false;
		}
		if (className.startsWith("javax.servlet") ||
				(className.startsWith("org.apache") &&
				!className.equals("org.apache.taglibs.standard.tag.common.sql.ResultImpl"))
				) {
			return true;
		}
		// else
		return false;
	}


	/**
	 *  Is this attribute name to be skipped when outputting the HTML data map.
	 *  <br>
	 *  If showAll, then don't skip any. Else skip attributes with invalid names
	 *  and attribute names which contain '.' except .comment attributes
	 *
	 *@param  attributeName  the attribute name.
	 *@return                true if the associated attribute value should be
	 *      skipped (not output)
	 */
	public boolean skipThisAttibute(String attributeName) {
		if (attributeName.trim().equals("")) {
			// this is handled separately
			return true;
		}
		if (showAll) {
			return false;
		}
		if (!WSTAttributeReader.attributeNamePattern.matcher(attributeName).matches() && (!showAll)) {
			return true;
			// skip bad names  let names with . in them through to next tests
		}
		if (attributeName.toLowerCase().endsWith(WST_COMMENT_ENDING)) {
			// always output .comment attributes
			return false;
		}
		if (attributeName.indexOf('.') >= 0) {
			// ?? . ?? cannot be accessed
			return true;
		}
		// else
		return false;
	}


	/**
	 *  Output an object in HTML data map format. <br>
	 *  Uses introspection on objects to extract attributes via get and is methods.
	 *
	 *@param  value            the object to output
	 *@param  indent           the current indent
	 *@param  level            the current level
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	private void objectToHTML(Object value, String indent, int level, Writer out) throws IOException {
		logger.entering(this.getClass().getName(), "objectToHTML", new Object[]{value, indent, new Integer(level)});
		// leave value wrapped for showFiters output, will show nestered filter wrapping.
		Object trueValue = value;
		
		IWSTFilter filter = null;
		if (value instanceof IWSTFilter) {
			while (value instanceof IWSTFilter) {
				// unwrap it first
				filter = (IWSTFilter) value;
				value = filter.getObject();
			}
		}
		
		level--;
		if (level < 0) {
			htmlDataObjectToString(trueValue, out);
			logger.exiting(this.getClass().getName(), "objectToHTML");
			return;
		}
		// else handle general object here
		if ((value instanceof String) || (value instanceof StringBuffer)
				 || (value instanceof Float) || (value instanceof Double)
				 || (value instanceof Integer) || (value instanceof Byte)
				 || (value instanceof Character)
				 || (value instanceof CharBuffer)
				 || (value instanceof Writer)
				 || (value instanceof Reader)
				 || (value instanceof OutputStream)
				 || (value instanceof InputStream)
				 || (value == null)) {
			htmlDataObjectToString(trueValue, out);
			logger.exiting(this.getClass().getName(), "objectToHTML");
			return;
		}

		// no need to check for [] of string, int, double
		// as have already called convert to array list before calling this method
		Class c = value.getClass();
		String className = c.getName();
		if (skipThisClass(className)) {
			out.write(className);
			logger.exiting(this.getClass().getName(), "objectToHTML");
			return;
		}


		Method methods[] = c.getMethods();
		// get the methods starting with getCap or isCap and with 0 parameters
		ArrayList methodsList = new ArrayList();
		ArrayList isMethodsList = new ArrayList();
		for (int i = 0; i < methods.length; i++) {
			if (!Modifier.isPublic(methods[i].getModifiers())) {
				// only list public methods
				continue;
			}
			if (methods[i].getParameterTypes().length > 0) {
				continue;
				// not ( )
			}
			String name = methods[i].getName();
			if (skipThisMethod(name)) {
				// skip these method names
				continue;
			}
			if (getPattern.matcher(name).matches()) {
				// matchs getXxx()
				methodsList.add(name);
			} else if (isPattern.matcher(name).matches()) {
				// matchs isXxx()
				isMethodsList.add(name);
			} else {
			}
		}

		if ((methodsList.size() <= 0) && (isMethodsList.size() <= 0)) {
			// no getXxx or isXxx methods for this class
			htmlDataObjectToString(trueValue, out);
			// just toString() it
			logger.exiting(this.getClass().getName(), "objectToHTML");
			return;
		}
		// else

		// now add isList to getList if get not already there
		String methodNames[] = (String[]) methodsList.toArray(new String[0]);
		for (int i = 0; i < isMethodsList.size(); i++) {
			String isName = (String) isMethodsList.get(i);
			// drop the is and add get
			String getName = "get" + isName.substring(2, isName.length());
			boolean found = false;
			for (int j = 0; j < methodNames.length; j++) {
				if (getName.equals(methodNames[j])) {
					found = true;
					break;
				}
			}
			if (!found) {
				methodsList.add(isName);
			}
		}
		// ok now have list of getXXX and isXXX methods
		
		// put in the plain toString() version between the <dd> and the <dl>
		// check if htmlDataObjectToString gives same value as ((Object)value).toString()
		// if so skip it, unless showFilters is on in which case show it to show the filter.
		try {
				String str = convertToString(trueValue).trim(); // let WSTFilters work on this
				if (showFilters || (!str.startsWith( value.getClass().getName()+"@") )) {
					htmlDataObjectToString(trueValue, out);
				}				
		} catch (Throwable t) {
				logger.log(Level.SEVERE, "Error converting " + value.getClass().getName() + " to String", t);
				out.write("Error converting " + value.getClass().getName() + " to String\n" + RobustFormatter.toString(t));
				//throw new RuntimeException("Error calling toString() on "+obj.getClass().getName(),t);
		}
		out.write(indent + "<DL>\n");

		StringBuffer nameBuf = null;
		for (int i = 0; i < methodsList.size(); i++) {
			String name = (String) methodsList.get(i);
			if (name.startsWith("get")) {
				nameBuf = new StringBuffer(name.substring(3, name.length()));
			} else if (name.startsWith("is")) {
				nameBuf = new StringBuffer(name.substring(2, name.length()));
			} else {
				// skip
				continue;
			}
			// make first char lc
			nameBuf.setCharAt(0, Character.toLowerCase(nameBuf.charAt(0)));
			String attrName = nameBuf.toString();
			out.write(indent + indentIncrement + "<DT>" + attrName + "</DT>\n");
			out.write(indent + indentIncrement + "<DD>");
			// call the method and add the result
			Object resultObj = null;
			try {
				Method m = c.getMethod(name, null);
				try {
					resultObj = m.invoke(value, null);
				} catch (Throwable tex) {
					resultObj = c.getName() + "." + name + "() \n throw exception " + tex.getClass().getName()
					 + ((tex.getMessage() == null)?"":(" "+tex.getMessage()));
					 Throwable cause = tex;
					 while((cause = cause.getCause()) != null) {
						 resultObj = ((String)resultObj) + "\n caused by "+ cause.getClass().getName()
						 + ((cause.getMessage() == null)?"":(" "+cause.getMessage()));
					}	 
				}

			} catch (Throwable t) {
				resultObj = "Error getting method " + name + "() on " + c.getName() +" \n throw exception " + t.getClass().getName()
				 + ((t.getMessage() == null)?"":(" "+t.getMessage()));
				 Throwable cause = t;
				 while((cause = cause.getCause()) != null) {
					 resultObj = ((String)resultObj) + "\n caused by "+ cause.getClass().getName()
					 + ((cause.getMessage() == null)?"":(" "+cause.getMessage()));
				}	 
				logger.log(Level.SEVERE, "Error calling " + name + "() on " + c.getName(), t);
				//throw new RuntimeException("Error calling "+name+"() on "+c.getName(),t);
			}
			// now have result process it as an object
			if ((resultObj instanceof Map) || (resultObj instanceof List) || (resultObj instanceof Enumeration)) {
				out.write("\n");
			}
			// re-wrapp it
			Object wrappedObj = wrapWithFilter(resultObj, filter);
			// only if filter not null
			if (level < 0) {
				htmlDataObjectToString(wrappedObj, out);
				// just toString it
			} else {
				processObjectHTML(wrappedObj, indent + indentIncrement, level, out);
			}
			if ((resultObj instanceof Map) || (resultObj instanceof List) || (resultObj instanceof Enumeration)) {
				out.write("\n");
				out.write(indent + indentIncrement + "</DD>\n");
			} else {
				out.write("</DD>\n");
			}
		}
		out.write(indent + "</DL>\n");
		logger.exiting(this.getClass().getName(), "objectToHTML");
		return;
		//return listToHTML((List)methodsList,indent,level++);  // actually the same level at this
	}


	/**
	 *  Comparator used to sort the attribute names in order
	 */
	protected Comparator keyComparator = new AttributeNameComparator();


	/**
	 *  Output this template, using the current attribute data map.<br>
	 *  If {@link WebStringTemplate#OUTPUT_DATA_DEFINITION_FLAG
	 *  wst.outputDataDefinition} is not false then instead of processing the
	 *  template, output the attributes as a HTML page.<p>
	 *
	 *
	 *
	 *@return    a String containing either the result of {@link #write} or {@link
	 *      #attributeHashtableToHTML attributeHashtableToHTML()}
	 */
	public String toString() {
		StringWriter buf = new StringWriter(8192);
		try {
			write(buf);
		} catch (IOException io) {
			warning("Got IOException writing to StringWriter????");
		}
		return buf.toString();
	}


	/**
	 *  The buffer size of the internal StringBuffer is controlled by the {@link
	 *  WebStringTemplate#OUTPUT_BUFFER_SIZE wst.buffersize} attribute or set to
	 *  8192 bytes if this attribute is missing or not an integer.
	 *
	 *@return    The outputBufferSize value
	 */
	public int getOutputBufferSize() {
		int bufferSize = DEFAULT_BUFFER_SIZE;
		if (attributes != null) {
			Object bufferSizeObj = get(OUTPUT_BUFFER_SIZE);
		  String bufferSizeStr = convertToString(bufferSizeObj);
			try {
				bufferSize = Integer.parseInt(bufferSizeStr);
			} catch (Throwable t) {
				// ignore
			}
		}
		return Math.max(0, bufferSize);
	}


	/**
	 *  Output this template, using the current attribute data map.<br>
	 *  If {@link WebStringTemplate#OUTPUT_DATA_DEFINITION_FLAG
	 *  wst.outputDataDefinition} is not false then instead of processing the
	 *  template, output the attributes as a HTML page.<p>
	 *
	 *
	 *
	 *@param  out              Description of the Parameter
	 *@exception  IOException  Description of the Exception
	 */
	public void write(Writer out) throws IOException {
		int maxDepth = DEFAULT_MAX_DEPTH;
		boolean writeAttributes = false;
		showAll = false;
		try {
			if (attributes != null) {

				Object writeAttributesObj = get(OUTPUT_DATA_DEFINITION_FLAG);
				writeAttributes = testObjectTrue(writeAttributesObj);

				if (writeAttributes) {

					Object showAllObj = get(OUTPUT_SHOW_ALL_FLAG);
					showAll = testObjectTrue(showAllObj);
					Object showFiltersObj = get(OUTPUT_SHOW_FILTERS_FLAG);
					showFilters = testObjectTrue(showFiltersObj);

					// check depth
					Object maxDepthObj = get(OUTPUT_MAX_DEPTH);
				  String maxDepthStr = convertToString(maxDepthObj);
					try {
						maxDepth = Integer.parseInt(maxDepthStr);
					} catch (Throwable t) {
						// ignore
					}
					maxDepth = Math.max(1, maxDepth);
				}
			}

			if (writeAttributes) {
				attributeHashtableToHTML(attributes, getName() + ".st", null, maxDepth, out);
			} else {
				super.write(out);
			}

		} finally {
			showAll = false;
			showFilters = false;
			// always reset show all and show filters
		}

	}

}

