/*
 *  Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsible for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package au.com.forward.webstringtemplate;

import au.com.forward.runWST.*;
import au.com.forward.runWST.maps.*;
import au.com.forward.runWST.wstAttributeReader.*;
import org.antlr.stringtemplate.*;
import au.com.forward.webstringtemplate.filters.*;

//import javax.servlet.*;
//import javax.servlet.http.*;
//import javax.servlet.jsp.PageContext;
//import javax.servlet.jsp.JspWriter;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.io.*;
import java.util.logging.*;

/**
 *  Manages a group of named mutually-referential WebStringTemplate objects.
 *  <br>
 *  See the documenation for details. <br>
 *  Currently the templates must all live under the same directory so that you
 *  can reference them as foo.st or gutter/header.st.<br>
 *  To refresh a group of templates, just create a new WebStringTemplateGroup
 *  and start pulling templates from there, or, set the refresh interval.<p>
 *
 *  Use getInstanceOf(template-name) to get a string template to fill in. The
 *  name of a template is the file name minus ".st" ending.
 *
 *@author     matthew ford
 *@created    December 14, 2003
 */
public class WebStringTemplateGroup extends StringTemplateGroup {

	/**
	 *  The Java logger for this class
	 */
	public final static Logger logger = Logger.getLogger("au.com.forward.webstringtemplate.WebStringTemplateGroup");

	/**
	 *  Maps data fileName to data hashtable
	 */
	private Hashtable data = new Hashtable();

	/**
	 *  Maps attribute extensions to filters
	 */
	private HashMap filters = new HashMap();

	private final static AttributeHashtable DATA_NOT_FOUND = new AttributeHashtable();

	/**
	 *  True if missing data should return an error path string, default false
	 */
	protected boolean missingDataErrorPathReturnFlag = false;


	/**
	 *  Create a StringTemplate of the correct class, used when sub-classing.
	 *
	 *@param  initialValues  map of attribute values
	 *@return                a stringtemplate of the correct class
	 */
	public StringTemplate createStringTemplate(Map initialValues) {
		return new WebStringTemplate(this, initialValues);
	}


	/**
	 *  Gets the missingDataErrorPathReturn setting.
	 *
	 *@return    true if missing data returns an error path, else false
	 */
	public boolean isMissingDataErrorPathReturn() {
		return missingDataErrorPathReturnFlag;
	}


	/**
	 *  Sets the missingDataErrorPathReturn
	 *
	 *@param  flag  true if missing data is to return an error path, else false
	 */
	public void setMissingDataErrorPathReturn(boolean flag) {
		missingDataErrorPathReturnFlag = flag;
	}


	/**
	 *  Set the default filter to use to filter strings passed to the template from
	 *  the data map. <br>
	 *  This filter is only used if the attribute is not already wrapped in a
	 *  filter.
	 *
	 *@param  filter                 The defaultFilter
	 *@return                        the previous defaultFilter or null if none
	 *      previously set.
	 *@throws  NullPointerException  if filter null.
	 */
	public IWSTFilter setDefaultFilter(IWSTFilter filter) {
		if (filter == null) {
			throw new NullPointerException("Default Filter cannot be null");
		}
		IWSTFilter returnFilter = (IWSTFilter) filters.get("");
		filters.put("", filter);
		return returnFilter;
	}


	/**
	 *  Gets the defaultFilter for this WebStringTemplateGroup
	 *
	 *@return    The defaultFilter
	 */
	public IWSTFilter getDefaultFilter() {
		return (IWSTFilter) filters.get("");
	}


	/**
	 *  Set the filter associated with this name extension. <br>
	 *  Don't set a filter for ".comment" as this extension is used for commenting
	 *  the data definition file.
	 *
	 *@param  nameExtension              The attribute extension to associate this
	 *      filter with. Any leading '.' is stripped.
	 *@param  filter                     The filter to associate with this
	 *      extension, null to remove association.
	 *@return                            the previous filter for this extension, or
	 *      null if none previously set.
	 *@throws  NullPointerException      if filter null and nameExtension is "".
	 *      I.e. cannot remove the default filter.
	 *@throws  IllegalArgumentException  if extension name is "comment" (ignoring
	 *      case).
	 */
	public IWSTFilter setFilter(String nameExtension, IWSTFilter filter) {
		if (nameExtension == null) {
			throw new NullPointerException("nameExtension cannot be null.");
		}
		nameExtension = nameExtension.trim();
		if (nameExtension.startsWith(".")) {
			nameExtension = nameExtension.substring(1);
		}
		if ((filter == null) && nameExtension.equals("")) {
			throw new IllegalArgumentException("Cannot remove the Default Filter.");
		}
		String comment = WebStringTemplate.WST_COMMENT_ENDING;
		if (comment.startsWith(".")) {
			comment = comment.substring(1);
		}
		if (nameExtension.equalsIgnoreCase(comment)) {
			throw new IllegalArgumentException("Cannot set a filter for '" + WebStringTemplate.WST_COMMENT_ENDING
					 + "' extensions.");
		}
		IWSTFilter returnFilter = null;
		returnFilter = (IWSTFilter) filters.get(nameExtension);
		if (filter == null) {
			filters.remove(nameExtension);
		} else {
			filters.put(nameExtension, filter);
		}
		return returnFilter;
	}


	/**
	 *  Gets the filter associated with this extension (case sensitive).<br>
	 *  Strips any leading '.' from the extension before doing lookup
	 *
	 *@param  nameExtension  Description of the Parameter
	 *@return                The filter assocaited with this extension or null if
	 *      none set.
	 */
	public IWSTFilter getFilter(String nameExtension) {
		if (nameExtension == null) {
			throw new NullPointerException("nameExtension cannot be null.");
		}
		nameExtension = nameExtension.trim();
		if (nameExtension.startsWith(".")) {
			nameExtension = nameExtension.substring(1);
		}
		return (IWSTFilter) filters.get(nameExtension);
	}


	/**
	 *  Returns a copy of the defined filters.<br>
	 *
	 *
	 *@return    Object[][2] of String key, IWSFilter pairs
	 */
	public Object[][] getAllFilters() {
		int noFilters = filters.size();
		Object[][] rtn = new Object[noFilters][2];
		int i = 0;
		Iterator it = filters.keySet().iterator();
		while (it.hasNext()) {
			String key = (String) it.next();
			rtn[i][0] = key;
			rtn[i][1] = filters.get(key);
			i++;
		}
		return rtn;
	}


	/**
	 *  The default error listener for all the WebStringTemplates
	 */
	public static StringTemplateErrorListener DEFAULT_SERVLET_ERROR_LISTENER =
		new StringTemplateErrorListener() {

			// this error method should never return
			public void error(String s, Throwable t) {
				// log it
				System.err.println(s);
				if (t != null) {
					t.printStackTrace(System.err);
				}
				//throw it
				throw new RunWSTException(s, t);
			}


			public void warning(String s) {
				// just log it
				System.out.println(s);
			}


			public void debug(String s) {
				//System.out.println(s);
			}
		};



	/**
	 *  Create a group manager for some templates, all of which are loaded from the
	 *  rootDir or classpath if rootDir is null. <br>
	 *  Uses "$[" and "]$" as delimiters.<p>
	 *
	 *  The default filter is WSTXmlFilter and the name extension .plain is set to
	 *  WSTPlainFiter and .xml is set to WSTXmlFilter
	 *
	 *@param  name     the name of group
	 *@param  rootDir  root dir to load templates from or null to use classpath.
	 *      Any trailing / or file.separator is stripped.
	 */
	public WebStringTemplateGroup(String name, String rootDir) {
		this(name, rootDir, "$[", "]$");
	}



	/**
	 *  Create a group manager for some templates, all of which are loaded from the
	 *  rootDir or classpath if rootDir is null. <br>
	 *  The default filter is WSTXmlFilter and the name extension .plain is set to
	 *  WSTPlainFiter and .xml is set to WSTXmlFilter
	 *
	 *@param  name            name of group
	 *@param  rootDir         root dir to load templates from or null to use
	 *      classpath. Any trailing / or file.separator is stripped
	 *@param  delimiterStart  start delimiter
	 *@param  delimiterStop   end delimiter
	 */
	public WebStringTemplateGroup(String name,
			String rootDir,
			String delimiterStart,
			String delimiterStop) {
		super(name, rootDir, delimiterStart, delimiterStop);
		if ((rootDir != null) && (rootDir.endsWith("/") || rootDir.endsWith("\\"))) {
			// strip it off as added later
			setRootDir(rootDir.substring(0, rootDir.length() - 1));
		}
		setErrorListener((StringTemplateErrorListener) DEFAULT_SERVLET_ERROR_LISTENER);
		data = new Hashtable();
		filters = new HashMap();
		// put in default filter and .xml and .plain
		IWSTFilter filter = new WSTXmlFilter();
		filters.put("", filter);
		filters.put("xml", filter);
		filters.put("plain", new WSTPlainFilter());
	}


	/**
	 *  Sets the default errorListener of the WebStringTemplates in this group
	 *
	 *@param  listener               The errorListener value
	 *@throws  NullPointerException  if listener is null.
	 */
	public void setErrorListener(StringTemplateErrorListener listener) {
		if (listener == null) {
			throw new NullPointerException("StringTemplateErrorListener cannot be null.");
		}
		super.setErrorListener(listener);
	}


	/**
	 *  Check if time to reload data and templates
	 */
	protected void checkRefreshInterval() {
		boolean timeToFlush =
				(refreshIntervalInSeconds == 0 || (System.currentTimeMillis() - lastCheckedDisk) >= refreshIntervalInSeconds * 1000);
		if (timeToFlush) {
			// throw away all data and templates
			data.clear();
			templates.clear();
			lastCheckedDisk = System.currentTimeMillis();
		}
	}


	/**
	 *  Gets the attributes from a HTML data file. <br>
	 *  Looks in the cache first and returns a clone of the previously loaded data
	 *  if it is available.<br>
	 *  Else if rootDir is null try and load as a resource from the classpath else
	 *  load from rootDir+'/'+fileName.<p>
	 *
	 *  If an error occures during loading, {@link
	 *  StringTemplateErrorListener#error} is called and the map set to null. Note:
	 *  the {@link StringTemplateErrorListener#error} may throw an exception and
	 *  not return. <p>
	 *
	 *  All attribute values are loaded as Strings. Each attribute is wrapped in a
	 *  WSTPlainFilter to ensure it is delivered to the template exactly as loaded.
	 *  <p>
	 *
	 *  To generate HTML data, create a template and set its attribute data map and
	 *  then set the attribute {@link WebStringTemplate#OUTPUT_DATA_DEFINITION_FLAG
	 *  wst.outputDataDefinition} to "true" in the template data map and call
	 *  template.toString() and save the resulting string in a file.
	 *
	 *@param  fileName               the name of the HTML data file to load
	 *      (including extension.)
	 *@return                        The template data map or null if error occured
	 *      in loading.
	 *@throws  NullPointerException  if fileName is null.
	 */
	public Map getData(String fileName) {
		if (fileName == null) {
			throw new NullPointerException("Data fileName cannot be null.");
		}
		if (getRootDir() != null) {
			fileName = getRootDir() + "/" + fileName;
		}
		checkRefreshInterval();
		// check if should reload data
		AttributeHashtable datatable = (AttributeHashtable) data.get(fileName);
		if (datatable != null) {
			return (AttributeHashtable) datatable.clone();
		}
		//else  don't look in supergroup for now

		WSTAttributeReader reader = new WSTAttributeReader(getErrorListener());
		try {
			if (getRootDir() == null) {
				datatable = reader.loadDataFromClasspath(fileName);
				// returns null on errors
			} else {
				datatable = reader.loadDataFromFile(fileName);
				// returns null on errors
			}
		} catch (RunWSTException rex) {
			// leave datatable as null and send this error to the System.out
			listener.error("Could not loading data " + fileName + " \n" + rex.getMessage() +
					((rex.getCause() == null) ? "" : "\n" + rex.getCause().getMessage()), rex);
		}
		if (datatable == null) {
			datatable = DATA_NOT_FOUND;
			// empty hashtable
		}
		data.put(fileName, datatable);
		return datatable;
	}



	/**
	 *  Gets the instanceOf a {@link WebStringTemplate} with an empty data map.
	 *  <br>
	 *  You need to cast the result i.e. (WebStringTemplate)getInstanceOf(name);<br>
	 *  Use {@link #getInstanceOf(String,String) getInstanceOf(name,(String)null)}; instead as this
	 *  returns a WebStringTemplate so a cast is not needed.
	 *
	 *@param  name  the file name of the template (without the .st)
	 *@return       the {@link WebStringTemplate} object
	 */
	public StringTemplate getInstanceOf(String name) {
		StringTemplate st = getInstanceOf(name, (HashMap) null);
		return st;
	}



	/**
	 *  Gets the instanceOf a {@link WebStringTemplate} loaded with data from a
	 *  HTML data file.
	 *
	 *@param  templateName  Name of template to load (without the trailing '.st')
	 *@param  dataFileName  the name of the data file to intialize this template
	 *      with (including extension) or null if no dataFile to be loaded.
	 *@return               The {@link WebStringTemplate} with attributes set to
	 *      those in the datafile
	 */
	public WebStringTemplate getInstanceOf(String templateName, String dataFileName) {
		WebStringTemplate wst = (WebStringTemplate) getInstanceOf(templateName);
		if (wst == null) {
			return wst;
		}
		if (dataFileName != null) {
			wst.setAttributes(getData(dataFileName));
		}
		return wst;
	}


	/**
	 *  Gets the instanceOf a {@link WebStringTemplate} loaded with data from a
	 *  map.
	 *
	 *@param  attributes  The new attributes value
	 *@param  wst         The new attributes value
	 *@return             The StringTemplate with attributes set to the response
	 *      attributes public StringTemplate getInstanceOf(String templateName,
	 *      ServletRequest request) { WebStringTemplate wst = (WebStringTemplate)
	 *      getInstanceOf(templateName); if (wst == null) { return wst; } / set
	 *      both the request attribute and the indivigual attributes return
	 *      setJSPAttributes(wst, request); }
	 */

	/**
	 *  Sets the parameters in the WebStringTemplate. <br>
	 *  After this call the parameters can be accessed using param.<i>name</i> .<p>
	 *
	 *  <b>NOTE:</b> <br>
	 *  After calling this method the post form data will no longer be available
	 *  for reading directly from the request object's input stream.
	 *
	 *@param  attributes  The new attributes value
	 *@param  wst         The new attributes value
	 *@return             The string template with the request parameters added
	 *      under param. .. public static Hashtable
	 *      getRequestAttributes(ServletRequest request) { /if (request instanceof
	 *      StringTemplateRequestWrapper) { / have attributes in map just return it
	 *      / return ((StringTemplateRequestWrapper) request).getAttributes(); /} /
	 *      else / initialize hash table to empty Hashtable table = new
	 *      Hashtable(); / load the attributes with those set in this request for
	 *      (Enumeration e = request.getAttributeNames(); e.hasMoreElements(); ) {
	 *      String name = (String) e.nextElement(); table.put(name,
	 *      request.getAttribute(name)); } return table; }
	 */

	/**
	 *  adds any attributes defined in this scope to the template The scopeObject
	 *  can be any of the predefined JSP objects pageContext, session, application,
	 *  response or config OR any object of one of those classes or an object of
	 *  class java.util.Map overwrites any existing attribute of the same name
	 *
	 *@param  attributes  The new attributes value
	 *@param  wst         The new attributes value
	 *@return             Description of the Return Value public static
	 *      StringTemplate setJSPAttributes(StringTemplate st, Object scopeObject)
	 *      { if (scopeObject == null) { return st; } / else if (scopeObject
	 *      instanceof PageContext) { / pageContext PageContext pageContext =
	 *      (PageContext) scopeObject; for (Enumeration e =
	 *      pageContext.getAttributeNamesInScope(PageContext.PAGE_SCOPE);
	 *      e.hasMoreElements(); ) { String name = (String) e.nextElement();
	 *      st.setAttribute(name, pageContext.getAttribute(name)); } } else if
	 *      (scopeObject instanceof HttpSession) { / session HttpSession
	 *      httpSession = (HttpSession) scopeObject; for (Enumeration e =
	 *      httpSession.getAttributeNames(); e.hasMoreElements(); ) { String name =
	 *      (String) e.nextElement(); st.setAttribute(name,
	 *      httpSession.getAttribute(name)); } } else if (scopeObject instanceof
	 *      ServletContext) { / application ServletContext servletContext =
	 *      (ServletContext) scopeObject; for (Enumeration e =
	 *      servletContext.getAttributeNames(); e.hasMoreElements(); ) { String
	 *      name = (String) e.nextElement(); st.setAttribute(name,
	 *      servletContext.getAttribute(name)); } } else if (scopeObject instanceof
	 *      ServletConfig) { / config ServletConfig servletConfig = (ServletConfig)
	 *      scopeObject; for (Enumeration e =
	 *      servletConfig.getInitParameterNames(); e.hasMoreElements(); ) { String
	 *      name = (String) e.nextElement(); st.setAttribute(name,
	 *      servletConfig.getInitParameter(name)); } } else if (scopeObject
	 *      instanceof ServletRequest) { / request ServletRequest request =
	 *      (ServletRequest) scopeObject; for (Enumeration e =
	 *      request.getAttributeNames(); e.hasMoreElements(); ) { String name =
	 *      (String) e.nextElement(); st.setAttribute(name,
	 *      request.getAttribute(name)); } } else if (scopeObject instanceof Map) {
	 *      / param header etc Map map = (Map) scopeObject; for (Iterator it =
	 *      map.keySet().iterator(); it.hasNext(); ) { Object nextName = it.next();
	 *      if (!(nextName instanceof String)) { continue; } String name = (String)
	 *      nextName; st.setAttribute(name, map.get(name)); } } else { throw new
	 *      RuntimeException("ScopeObject " + scopeObject.getClass().getName() + "
	 *      not a valid class type."); } return st; }
	 */

	/**
	 *  Add these attributes to the top level of the data map of the template. <br>
	 *  If any already exist they are overwritten by these. <br>
	 *  Any non-String keys are quietly ignored and any invalid attribute names are
	 *  not loaded.<br>
	 *
	 *
	 *@param  wst         the WebStringTemplate
	 *@param  attributes  the map containing the attributes to add
	 *@return             The updated string template
	 */
	public static WebStringTemplate setAttributes(WebStringTemplate wst, Map attributes) {
		for (Iterator it = attributes.keySet().iterator(); it.hasNext(); ) {
			Object key = it.next();
			if (key == null) {
				continue;
			}
			if (!(key instanceof String)) {
				continue;
			}
			String name = (String) key;
			try {
				wst.setAttribute(name, attributes.get(name));
			} catch (Throwable t) {
				// ignore
			}
		}
		return wst;
	}

}

