/*
 *  Copyright (c) 2003 Forward Computing and Control Pty. Ltd.
 *  NSW. Australia,  www.forward.com.au
 *  All rights reserved.
 *  Redistribution of the source of this package, with or without
 *  modification, is NOT permitted.
 *  Redistribution of the compiled java .class files of this package, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions in compiled java .class form must reproduce the above
 *  copyright notice, this list of conditions and the following disclaimer
 *  in the documentation and/or other materials provided with the distribution.
 *  2. If the software is modified, the compiled java .class distribution must
 *  include, in the documentation and/or other materials provided with the
 *  distribution, a statement that the software has been modified and
 *  identify the person who was responsible for the modifications.
 *  3. Neither the name of the author nor Forward Computing and Control may be
 *  used to endorse or promote products derived from this software
 *  without specific prior written permission.
 *  4.  All advertising materials mentioning features or use of this
 *  software must display the following acknowledgment:
 *  "This product includes software developed by Matthew Ford
 *  and Forward Computing and Control Pty. Ltd.
 *  (http://www.forward.com.au)."
 *  THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 *  IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 *  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 *  NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 *  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 *  THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 *  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 *  THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package au.com.forward.webstringtemplate;

import au.com.forward.runWST.*;
import au.com.forward.runWST.maps.*;
import au.com.forward.runWST.wstAttributeReader.*;
import org.antlr.stringtemplate.*;
import au.com.forward.webstringtemplate.filters.*;

import javax.servlet.*;
import javax.servlet.http.*;
import javax.servlet.jsp.PageContext;
import javax.servlet.jsp.JspWriter;

import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Map;
import java.util.HashMap;
import java.util.Iterator;
import java.io.*;
import java.util.logging.*;

/**
 *  This class provides a set of useful server side WebStringTemplate routines.
 *  <br>
 *  These are in a separate class so that RunWST does not need any
 *  javax.servlet.* classes in order to run.
 *
 *@author     matthew ford
 *@created    December 14, 2003
 */
public class WebStringTemplateServer {

	/**
	 *  The Java logger for this class
	 */
	public final static Logger logger = Logger.getLogger("au.com.forward.webstringtemplate.WebStringTemplateServer");


	/**
	 *  Write out this template using the attributes in the pageContext.<br>
	 *
	 *
	 *@param  wst              the template to write
	 *@param  pageContext      the JSP page context containing the attributes to be
	 *      inserted in the template
	 *@exception  IOException  Description of the Exception
	 */
	public static void writePageTemplate(WebStringTemplate wst, PageContext pageContext) throws IOException {
		JspWriter out = pageContext.getOut();
		// pickup all these attributes
		setJSPAttributes(wst, pageContext);
		// remove any text generated by this page so far mainly blank lines
		//out.clear();  no leave any output generated so far so that preludes take effect.
		out.write(wst.toString());
	}


	/**
	 *  Get a template loaded with default data, using cache if possiable. <br>
	 *  This method looks for a default data file by adding <b>.data</b> to the
	 *  templateName and loads the data if found.
	 *
	 *@param  application   the JSP application object
	 *@param  groupName     the name of the group, used to look for an existing
	 *      group.
	 *@param  groupRootDir  the rootDir of the group, only used in a new group is
	 *      created.
	 *@param  templateName  the template to load (without any extension)
	 *@return               the template, loaded with data from template.data if
	 *      the data file was found.
	 */
	public static WebStringTemplate getInstanceOf(ServletContext application, String groupName,
			String groupRootDir, String templateName) {
		return getInstanceOf(application, groupName, groupRootDir, templateName, templateName + ".data");
	}


	/**
	 *  Set the refresh interval for this application group.<br>
	 *  If this group name does not exist in the application cache an exception is
	 *  thrown.
	 *
	 *@param  application             the JSP application object
	 *@param  groupName               the name of the group, used to look for an
	 *      existing group.
	 *@param  refreshInterval         the refresh interval (Sec), if &lt; 0 set to
	 *      0, 0 means no caching.
	 *@throws  IllegalStateException  if group does not already exists in this
	 *      application.
	 */
	public static void setRefreshInterval(ServletContext application, String groupName,
			int refreshInterval) {
		if (refreshInterval < 0) {
			refreshInterval = 0;
		}
		WebStringTemplateGroup group = getGroupIfExists(application, groupName);
		if (group == null) {
			throw new IllegalStateException("WebStringTemplateGroup '" + groupName + "' does not exist in this application.");
		}
		group.setRefreshInterval(refreshInterval);
	}


	/**
	 *  Get a template loaded with default data, using cache if possiable. <br>
	 *  This method looks for the specified default data file and loads the data if
	 *  found.
	 *
	 *@param  application   the JSP application object
	 *@param  groupName     the name of the group, used to look for an existing
	 *      group.
	 *@param  groupRootDir  the rootDir of the group, only used in a new group is
	 *      created.
	 *@param  templateName  the template to load (without any extension)
	 *@param  dataFileName  the dataFile to load (with extension) or null if no
	 *      dataFile to be loaded
	 *@return               the template, loaded with data, if the data file was
	 *      found.
	 */
	public static WebStringTemplate getInstanceOf(ServletContext application, String groupName,
			String groupRootDir, String templateName, String dataFileName) {
		WebStringTemplateGroup group = getInstanceOfGroup(application, groupName, groupRootDir);
		if (group == null) {
			return null;
		}
		// else
		return group.getInstanceOf(templateName, dataFileName);
	}


	/**
	 *  Get a group from this ServletContext. <br>
	 *  Creates a new group in this application if none found.
	 *
	 *@param  application   the JSP application object
	 *@param  groupName     the name of the group, used to look for an existing
	 *      group.
	 *@param  groupRootDir  the rootDir of the group, only used in a new group is
	 *      created.
	 *@return               the existing group or a new group if none existed.
	 */
	public static WebStringTemplateGroup getInstanceOfGroup(ServletContext application, String groupName,
			String groupRootDir) {
		WebStringTemplateGroup group = null;
		synchronized (application) {
			Hashtable grouptable = (Hashtable) application.getAttribute("WebStringTemplateGroupTable");
			if (grouptable == null) {
				grouptable = new Hashtable();
			}
			application.setAttribute("WebStringTempleteGroupTable", grouptable);
			group = (WebStringTemplateGroup) grouptable.get(groupName);
			if (group == null) {
				group = new WebStringTemplateGroup(groupName, groupRootDir);
				grouptable.put(groupName, group);
			}
		}
		return group;
	}


	/**
	 *  Gets the groupIfExists attribute of the WebStringTemplateServer class
	 *
	 *@param  application  Description of the Parameter
	 *@param  groupName    Description of the Parameter
	 *@return              The groupIfExists value
	 */
	private static WebStringTemplateGroup getGroupIfExists(ServletContext application, String groupName) {
		WebStringTemplateGroup group = null;
		synchronized (application) {
			Hashtable grouptable = (Hashtable) application.getAttribute("WebStringTemplateGroupTable");
			if (grouptable == null) {
				return null;
			}
			return (WebStringTemplateGroup) grouptable.get(groupName);
		}
	}


	/**
	 *  Adds any attributes defined in the given scope to the template. <br>
	 *  The scopeObject can be any of the predefined JSP objects<br>
	 *  pageContext, session, application, response or config, or one of those
	 *  objects' classes <br>
	 *  OR<br>
	 *  an object of class java.util.Map<p>
	 *
	 *  Any existing attributes of the same name are overwritten.
	 *
	 *@param  wst          the template to load with the JSP attributes
	 *@param  scopeObject  the scope object to load from or a Map object
	 *@return              the template loaded with these attributes
	 */
	public static WebStringTemplate setJSPAttributes(WebStringTemplate wst, Object scopeObject) {
		if (scopeObject == null) {
			return wst;
		}
		// else
		if (scopeObject instanceof PageContext) {
			// pageContext
			PageContext pageContext = (PageContext) scopeObject;
			for (Enumeration e = pageContext.getAttributeNamesInScope(PageContext.PAGE_SCOPE);
					e.hasMoreElements(); ) {
				String name = (String) e.nextElement();
				wst.setAttribute(name, pageContext.getAttribute(name));
			}
		} else if (scopeObject instanceof HttpSession) {
			// session
			HttpSession httpSession = (HttpSession) scopeObject;
			for (Enumeration e = httpSession.getAttributeNames(); e.hasMoreElements(); ) {
				String name = (String) e.nextElement();
				wst.setAttribute(name, httpSession.getAttribute(name));
			}
		} else if (scopeObject instanceof ServletContext) {
			// application
			ServletContext servletContext = (ServletContext) scopeObject;
			for (Enumeration e = servletContext.getAttributeNames(); e.hasMoreElements(); ) {
				String name = (String) e.nextElement();
				wst.setAttribute(name, servletContext.getAttribute(name));
			}
		} else if (scopeObject instanceof ServletConfig) {
			// config
			ServletConfig servletConfig = (ServletConfig) scopeObject;
			for (Enumeration e = servletConfig.getInitParameterNames(); e.hasMoreElements(); ) {
				String name = (String) e.nextElement();
				wst.setAttribute(name, servletConfig.getInitParameter(name));
			}
		} else if (scopeObject instanceof ServletRequest) {
			// request
			ServletRequest request = (ServletRequest) scopeObject;
			for (Enumeration e = request.getAttributeNames(); e.hasMoreElements(); ) {
				String name = (String) e.nextElement();
				wst.setAttribute(name, request.getAttribute(name));
			}
		} else if (scopeObject instanceof Map) {
			// param header etc
			Map map = (Map) scopeObject;
			for (Iterator it = map.keySet().iterator(); it.hasNext(); ) {
				Object nextName = it.next();
				if (!(nextName instanceof String)) {
					continue;
				}
				String name = (String) nextName;
				wst.setAttribute(name, map.get(name));
			}
		} else {
			throw new RuntimeException("ScopeObject " + scopeObject.getClass().getName() +
					" not a valid class type.");
		}
		return wst;
	}

}

