/*
 Modified by Matthew Ford 21st Dec. 2003
 modifications marked by // mod mpf
 
 [The "BSD licence"]
 Copyright (c) 2003 Terence Parr
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions
 are met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in the
    documentation and/or other materials provided with the distribution.
 3. The name of the author may not be used to endorse or promote products
    derived from this software without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
package org.antlr.stringtemplate.language;

import org.antlr.stringtemplate.StringTemplate;
import org.antlr.stringtemplate.StringTemplateGroup;
import org.antlr.stringtemplate.language.Expr;
import org.antlr.stringtemplate.language.StringTemplateAST;
import org.antlr.stringtemplate.language.StringTemplateLanguageEvaluator;

import au.com.forward.logging.*; // mod mpf
import java.util.logging.*; // mod mpf
import au.com.forward.webstringtemplate.filters.*; // mod mpf
import au.com.forward.webstringtemplate.*; // mod mpf

import java.io.Writer;
import java.io.IOException;
import java.io.StringWriter;
import java.util.*;
import java.lang.reflect.Method;

import antlr.collections.AST;
import antlr.RecognitionException;

/** A single string template expression enclosed in $...; separator=...$
 *  parsed into an AST chunk to be evaluated.
 */
public class ASTExpr extends Expr {
    public static final String DEFAULT_ATTRIBUTE_NAME = "attr";
    public static final String DEFAULT_INDEX_VARIABLE_NAME = "list_i"; // mod mpf
    public static final String DEFAULT_INDEX_FIRST_NAME = "list_first"; // mod mpf
    public static final String DEFAULT_INDEX_LAST_NAME = "list_last"; // mod mpf

	AST exprTree = null;

	/**
	 *  Description of the Field
	 *  // mod mpf
	 */
	public final static Logger logger = Logger.getLogger("org.antlr.stringtemplate.language.ASTExpr");
	
    /** store separator etc... */
    Map options = null;

    public ASTExpr(StringTemplate enclosingTemplate, AST exprTree, Map options) {
		super(enclosingTemplate);
        this.exprTree = exprTree;
        this.options = options;
    }

	/** Return the tree interpreted when this template is written out. */
	public AST getAST() {
		return exprTree;
	}

    /** To write out the value of an ASTExpr, invoke the evaluator in eval.g
     *  to walk the tree writing out the values.  For efficiency, don't
     *  compute a bunch of strings and then pack them together.  Write out directly.
     */
    public void write(StringTemplate self, Writer out) throws IOException {
        if ( exprTree==null || self==null || out==null ) {
            return;
        }
        //System.out.println("evaluating tree: "+exprTree.toStringList());
        StringTemplateLanguageEvaluator eval =
                new StringTemplateLanguageEvaluator(self,this,out);
        try {
            eval.action(exprTree); // eval and write out tree
        }
        catch (RecognitionException re) {
            self.error("can't evaluate tree: "+exprTree.toStringList(), re);
        }
    }

    // HELP ROUTINES CALLED BY EVALUATOR TREE WALKER

    public Object applyListOfAlternatingTemplates(StringTemplate self,
                                                  Object attributeValue,
                                                  List templatesToApply)
    {
        if ( attributeValue==null || templatesToApply==null || templatesToApply.size()==0 ) {
            return null; // do not apply if missing templates or empty value
        }
        //System.out.println("list of templates: "+templatesToApply);
        StringTemplate embedded = null;
        Map argumentContext = null;
				
				String propertyPath = "";
				IWSTFilter filter = null;
				while (attributeValue instanceof IWSTFilter) {
					filter = (IWSTFilter)attributeValue;
					attributeValue = filter.getObject();
					propertyPath = filter.getPath();
				}
				
				if (attributeValue == null) {
					return null;
				}
        
 
        if ( attributeValue instanceof List ) {
            List attributeList = (List)attributeValue;
            List resultVector = new ArrayList(attributeList.size());
            for (int i=0; i< attributeList.size(); i++) {
                Object ithValue = attributeList.get(i);
                if ( ithValue==null ) {
                    // weird...a null value in the list; make "" and continue
                    //continue;
										ithValue="";  // mod mpf
                }
								
                int templateIndex = i % templatesToApply.size(); // rotate through
                embedded = (StringTemplate)templatesToApply.get(templateIndex);
                argumentContext = new HashMap();
								
								if (self instanceof WebStringTemplate) {
									ithValue = ((WebStringTemplate)self).wrapWithFilter(ithValue,filter);
									((IWSTFilter)ithValue).setPath(propertyPath+"."+(i+1));
								}	
								
                argumentContext.put(DEFAULT_ATTRIBUTE_NAME, ithValue);
                argumentContext.put(DEFAULT_INDEX_VARIABLE_NAME, new Integer(i+1)); // mod mpf
                argumentContext.put(DEFAULT_INDEX_FIRST_NAME, (i==0)?"true":"false"); // mod mpf
                argumentContext.put(DEFAULT_INDEX_LAST_NAME, (i==(attributeList.size()-1))?"true":"false" ); // mod mpf
                embedded.setArgumentContext(argumentContext);
                evaluateArguments(embedded);
                /*
                System.out.println("applyTemplate("+embedded.getName()+", args="+argumentContext+
                        " to attribute value "+attributeValue);
                */
                Object value = embedded.toString();  // apply template
                resultVector.add(value);
                // can try to save embedded string template one day instead of
                // doing a toString().
            }
            //return resultVector;
						IWSTFilter rtnFilter = new WSTPlainFilter(resultVector); // mod mpf
						rtnFilter.setPath(propertyPath); // mod mpf
						return rtnFilter; // mod mpf
        }
        else {
            /*
            System.out.println("setting attribute "+DEFAULT_ATTRIBUTE_NAME+" in arg context of "+
            embedded.getName()+
            " to "+attributeValue);
            */
            embedded = (StringTemplate)templatesToApply.get(0);
            argumentContext = new HashMap();
						if (self instanceof WebStringTemplate) {
							attributeValue = ((WebStringTemplate)self).wrapWithFilter(attributeValue,filter);
							((IWSTFilter)attributeValue).setPath(propertyPath);
						}	
            argumentContext.put(DEFAULT_ATTRIBUTE_NAME, attributeValue);
            argumentContext.put(DEFAULT_INDEX_VARIABLE_NAME, new Integer(1));
            argumentContext.put(DEFAULT_INDEX_FIRST_NAME, "true"); // mod mpf
            argumentContext.put(DEFAULT_INDEX_LAST_NAME, "true"); // mod mpf
            embedded.setArgumentContext(argumentContext);
            evaluateArguments(embedded);
            //return embedded;
						IWSTFilter rtnFilter = new WSTPlainFilter(embedded); // mod mpf
						rtnFilter.setPath(propertyPath); // mod mpf
						return rtnFilter; // mod mpf
        }
    }

		/* ***************** this method replaced with one that gets isXxx() if no getXxx() ******* mod mpf */
    /** Return o.getPropertyName() given o and propertyName.
     *  If the method returns a boolean, I don't do anything special.  This
	 *  means that IF statements will always find a value and hence $if(boolean)$
	 *  is always true.
     *
    public Object getObjectProperty(StringTemplate self, Object o, String propertyName) {
        if ( o==null || propertyName==null ) {
            return null;
        }
        Class c = o.getClass();
        String methodName = "get"+
                Character.toUpperCase(propertyName.charAt(0))+
                propertyName.substring(1,propertyName.length());
        Object value = null;
        try {
            Method m = c.getMethod(methodName, null);
            value = m.invoke(o,null);
            if ( StringTemplate.isDebugMode() ) self.debug(methodName+"() returns "+value+"<"+value.getClass().getName()+">");
        }
        catch (Exception e) {
            self.error("Can't get property "+propertyName+" using method "+methodName+
                       " from "+c.getName()+" instance.", e);
        }
        return value;
    }
		
		*/

		/**
		 *  returns a String error message wrapped in a filter with its path set.
		 *  Used for returning $[path]$ when data not found.
		 *  Returns null wrapped in a filter if errorPathReturns disabled at the group level.
		 *
		 *@param path the actual path of this data
		 *@param pathMsg what is put inside the $[  ]$ return string, may inlcude a method call error message
		 *@param filter the original filter for this object, not used. Replaced with WSTMissingFilter
		 *@return an error path return string wrapped in a WSTMissingFilter.
		 *
		 */
		protected IWSTFilter setErrorReturn(WebStringTemplate self, String path, String pathMsg, IWSTFilter filter) {
			String errorRtnStr = null;
			if ( ((WebStringTemplateGroup)self.getGroup()).isMissingDataErrorPathReturn() ) {
				
				errorRtnStr = self.getGroup().getDelimiterStart() +pathMsg
						   					+self.getGroup().getDelimiterStop();
			}									
  		IWSTFilter errorRtn = new WSTMissingFilter(errorRtnStr); 
			errorRtn.setPath(path);
			return errorRtn;
		}	
		
    /** Return o.getPropertyName() given o and propertyName.  
		 *  converts results to ArrayList if possiable
		 *  Looks for isXxx() method if getXxx() not found
		 *  Returns $[attributePath]$ if not found
     */
    public Object getObjectProperty(StringTemplate self, Object o, String propertyName) {
        if (propertyName==null) {  // o == null handled below
            return null;
        }
        Object value = null;
				String propertyPath = "";
				IWSTFilter filter = null;
				while (o instanceof IWSTFilter) {
					filter = (IWSTFilter)o;
					o = filter.getObject();
					propertyPath = filter.getPath();
				}
				if (propertyPath.equals("")) {
					propertyPath = propertyName;
				} else {
					propertyPath += "."+propertyName;
				}
				if (o == null) {
					// value == null
					self.warning("Could not find '"+propertyPath+"' in instance of\n"+
						   "Template '"+self.getName()+"'");
					// convert value if array
					if (self instanceof WebStringTemplate) {
						value = setErrorReturn((WebStringTemplate)self,propertyPath,propertyPath,filter);
					}	
					return  value; // self.convertToArrayList(value);
				}				 
        Class c = o.getClass();
				if (o instanceof Map) {
					  java.util.Map map = (java.util.Map)o; 
					if (map.containsKey(propertyName)) {
						value = map.get(propertyName);
					} else {
						self.warning("Could not find '"+propertyPath+"' in instance of\n"+
							 "Template '"+self.getName()+"'");
						if (self instanceof WebStringTemplate) {
							value = setErrorReturn((WebStringTemplate)self,propertyPath,propertyPath,filter);
						}	
						return value;
					}
					
				} else {	
					// assume it is a bean
					// look for get then is methods
					String methodName =	Character.toUpperCase(propertyName.charAt(0))+
									propertyName.substring(1,propertyName.length());
					String callMethodName = "get"+methodName;
					Method m = null;
					try {
							m = c.getMethod(callMethodName, null);
					}	catch (Exception e) {
						// ignore if no getString method
					}
					if (m == null) {
						// try isXXX()
						callMethodName = "is"+methodName;
						try {
								m = c.getMethod(callMethodName, null);
						}	catch (Exception e) {
							// ignore if no isString method
						}
					}
					if (m!=null) {
						try {
							value = m.invoke(o,null);
						} catch (Throwable t) {
								self.error("Could not retrieve property '"+propertyPath+"'. "+
								  "Error occured executing '"+callMethodName+"()'"
								+" in "+c.getName()+" instance in\n Template '"+self.getName()+"'\n",t);
							if (self instanceof WebStringTemplate) {
								value = setErrorReturn((WebStringTemplate)self,propertyPath,
														propertyPath+".Error_"+callMethodName,filter);
								return value;	 
							}	
						}							 
					} else {
						// equals null method not found
							self.warning("Could not get property '"+propertyPath+"' using "
							+"'get"+methodName+"()' or "
							+"'is"+methodName+"()'"
							+" from "+c.getName()+" instance in\n Template '"+self.getName()+"'");
						if (self instanceof WebStringTemplate) {
							value = setErrorReturn((WebStringTemplate)self,propertyPath,propertyPath,filter);
							return value; 
						}	
					}							 
				}	
					// convert value if array
				if (self instanceof WebStringTemplate) {
					value = ((WebStringTemplate)self).wrapWithFilter(value,filter);
					((IWSTFilter)value).setPath(propertyPath);
				}	
				return  self.convertToArrayList(value);  // adds default filter is this is not already filtered					
    }

    /** For strings or other objects, catenate and return.
     */
    public Object add(Object a, Object b) {
        if ( a==null ) { // a null value means don't do cat, just return other value
            return new WSTPlainFilter(""+b);    // mod mpf  was b
        }
        else if ( b==null ) {
            return new WSTPlainFilter(""+a);    // mod mpf  was a
        }
        return new WSTPlainFilter(getEnclosingTemplate().convertToString(a) 
				    + getEnclosingTemplate().convertToString(a));
						 // mod mpf .toString() + b.toString();
    }

    /** Call a string template with args and get text result */
    public Object getTemplateText(StringTemplate self,   // mod mpf was String
                                  String templateName,
                                  Map argumentContext)
    {
        StringTemplateGroup group = self.getGroup();
        StringTemplate embedded = group.getEmbeddedInstanceOf(self, templateName);
        if ( embedded==null ) {
            self.error("cannot make emedded instance of "+templateName);
            return null;
        }
        embedded.setArgumentContext(argumentContext);
        String text = embedded.toString();
        //return text;
				return new WSTPlainFilter(text); // mod mpf
    }

    /** How to spit out an object.  If it's not a StringTemplate nor a
     *  List, just do o.toString().  If it's a StringTemplate,
     *  do o.write(out).  If it's a Vector, do a write(out,
     *  o.elementAt(i)) for all elements.  Note that if you do
     *  something weird like set the values of a multivalued tag
     *  to be vectors, it will effectively flatten it.
     *
     *  If self is an embedded template, you might have specified
     *  a separator arg; used when is a vector.
     */
    public void writeAttribute(StringTemplate self, Object o, Writer out) {
        Object separator = null;
        if ( options!=null ) {
            separator = options.get("separator");
        }
        write(self,o,out,separator);
    }

		/* **************** replaced by write below mod mpf **************
    protected void write(StringTemplate self, Object o, Writer out, Object separator) {
        if ( o==null ) {
            return;
        }
        try {
            if ( o instanceof StringTemplate ) {
                ((StringTemplate)o).write(out);
            }
            else if ( o instanceof List ) {
                Iterator e = ((List)o).iterator();
                String separatorString = null;
                if ( separator!=null ) {
                    separatorString = computeSeparator(self, separator);
                }
                while ( e.hasNext() ) {
                    write(self, e.next(), out, separator);
                    if ( e.hasNext() ) {
                        if ( separator!=null ) {
                            out.write(separatorString);
                        }
                    }
                }
            }
            else {
                out.write(getEnclosingTemplate().convertToString(o)); // mod mpf was o.toString());
            }
        }
        catch (Throwable t) { //was IOException io) {  mod mpf
            self.error("problem writing object: "+getEnclosingTemplate().convertToString(o), t); // mod mpf
        }
    }
    **************************** */
		
    protected void write(StringTemplate self, Object o, Writer out, Object separator) {
        if ( o==null ) {
            return;
        }
				Object orgObj = o;              // mod mpf
				IWSTFilter filter = null;       // mod mpf
				while (o instanceof IWSTFilter) {  // mod mpf
					filter = (IWSTFilter)o;       // mod mpf
					o = filter.getObject();       // mod mpf
				}                               // mod mpf
				if (o == null) {                // mod mpf
					return;                       // mod mpf
				}                               // mod mpf   
				
        try {
            if ( o instanceof StringTemplate ) {
                ((StringTemplate)o).write(out);
            }
            else if ( o instanceof List ) {
                Iterator e = ((List)o).iterator();
                String separatorString = null;
                if ( separator!=null ) {
                    separatorString = computeSeparator(self, separator);
                }
                while ( e.hasNext() ) {
									  Object nextObj = e.next();  // mod mpf
										if (self instanceof WebStringTemplate) {  // mod mpf
											nextObj = ((WebStringTemplate)self).wrapWithFilter(nextObj,filter); // mod mpf
										}	// mod mpf
                    write(self, nextObj, out, separator);  // mod mpf
                    if ( e.hasNext() ) {
                        if ( separator!=null ) {
                            out.write(separatorString);
                        }
                    }
                }
            }
            else {
                out.write(self.convertToString(orgObj)); // mod mpf was o.toString());
            }
        }
        catch (Throwable t) { //was IOException io) {  mod mpf
            self.error("problem writing object: "+getEnclosingTemplate().convertToString(o), t); // mod mpf
        }
    }
		
		
    /** A separator is normally just a string literal, but is still an AST that
     *  we must evaluate.  The separator can be any expression such as a template
     *  include or string cat expression etc...
     */
    protected String computeSeparator(StringTemplate self, Object separator) {
//System.out.println("called computeSeparator with "+separator.getClass().getName()+" "+separator);			
        if ( separator==null ) {
            return null;
        }
        if ( separator instanceof StringTemplateAST ) {
            StringTemplateAST separatorTree = (StringTemplateAST)separator;
            // must evaluate, writing to a string so we can hand on to it
            ASTExpr e = new ASTExpr(getEnclosingTemplate(),separatorTree,null);
            StringWriter sw = new StringWriter();
            try {
                e.write(self,sw);
            }
            catch (IOException ioe) {
                self.error("can't evaluate separator expression", ioe);
            }
            return sw.toString();
        }
        else {
            // just in case we expand in the future and it's something else
            return getEnclosingTemplate().convertToString(separator); // mod mpf was .toString();
        }
    }

    protected void evaluateArguments(StringTemplate self) {
        StringTemplateAST argumentsAST = self.getArgumentsAST();
        if ( argumentsAST==null || argumentsAST.getFirstChild()==null ) {
            // return immediately if missing tree or no actual args
            return;
        }
        StringTemplateLanguageEvaluator eval =
                new StringTemplateLanguageEvaluator(self,this,null);
        try {
            // using any initial argument context (such as when attr is set),
            // evaluage the arg list like bold(item=attr).  Since we pass
            // in any existing arg context, that context gets filled with
            // new values.  With bold(item=attr), context becomes:
            // {[attr=...],[item=...]}.
            Map ac = eval.argList(argumentsAST, self.getArgumentContext());
            self.setArgumentContext(ac);
        }
        catch (RecognitionException re) {
            self.error("can't evaluate tree: "+argumentsAST.toStringList(), re);
        }
    }
}
